/*
 * Copyright (c) 2015, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <avr/io.h>
#include <ucg_uart.h>
#include <ucg_avr_uart.h>

static void disable_tx_irq(struct ucg_uart *uart)
{
	struct ucg_avr_uart *avr_uart = uart->driver_data;
	*avr_uart->reg_ucsrb &= ~(1 << avr_uart->bit_udrie);
}

static void enable_tx_irq(struct ucg_uart *uart)
{
	struct ucg_avr_uart *avr_uart = uart->driver_data;
	*avr_uart->reg_ucsrb |= (1 << avr_uart->bit_udrie);
}

static uint8_t tx_ready(struct ucg_uart *uart)
{
	struct ucg_avr_uart *avr_uart = uart->driver_data;
	return !!(*avr_uart->reg_ucsra & (1 << avr_uart->bit_udre));
}

static void set_udr(struct ucg_uart *uart, char c)
{
	struct ucg_avr_uart *avr_uart = uart->driver_data;
	*avr_uart->reg_udr = c;
}

static char get_udr(struct ucg_uart *uart)
{
	struct ucg_avr_uart *avr_uart = uart->driver_data;
	return *avr_uart->reg_udr;
}

static int setconf(struct ucg_uart *uart, const struct ucg_uart_config *conf)
{
	struct ucg_avr_uart *avr_uart = uart->driver_data;
	uint16_t baudreg;
	uint8_t lo, hi;

	if (conf->enable == 0) {
		*avr_uart->reg_ucsrb = 0;
		return 0;
	}

	/* set msb bit of hi to 0 (useful fot uC with URSEL, and not
	 * important for the others because the baudreg will never be
	 * as big */
	baudreg = (F_CPU / (conf->baudrate * 8UL)) - 1;
	lo = (uint8_t)baudreg;
	hi = (uint8_t)((baudreg >> 8) & 0x7F);

	*avr_uart->reg_ubrrl = lo;
	*avr_uart->reg_ubrrh = hi;
	*avr_uart->reg_ucsrb = (
		(1 << avr_uart->bit_txen) |
		(1 << avr_uart->bit_rxen) |
		(1 << avr_uart->bit_rxcie));

	return 0;
}

const struct ucg_uart_driver_ops avr_uart_ops = {
	.disable_tx_irq = disable_tx_irq,
	.enable_tx_irq = enable_tx_irq,
	.tx_ready = tx_ready,
	.set_udr = set_udr,
	.get_udr = get_udr,
	.setconf = setconf,
};
