/*
 * Copyright (c) 2005-2015, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <uart.h>
#include <uart_defs.h>
#include <uart_private.h>

#define UART_HW_NUM 1

struct cirbuf g_tx_fifo[UART_HW_NUM];
struct cirbuf g_rx_fifo[UART_HW_NUM];

/** this structure stores the configuration of the uart */
struct uart_config {
      uint8_t enabled          : 1, /**< enable or disable the uart */
      intr_enabled     : 1, /**< use interruptions or not */
      use_double_speed : 1, /**< less acurate, but can reach faster baudrate */
      parity           : 2, /**< none, odd or even */
      stop_bits        : 1, /**< 1 or 2 bits at the end of the frame */
      reserved         : 1; /**< nothing for now */
      uint8_t nbits;                /**< number of bits in frame, 5,6,7,8 or 9 */
      uint32_t baudrate;             /**< speed of uart */
};

/** The emission fifo of uart */
extern struct cirbuf g_tx_fifo[UART_HW_NUM];

/** The reception fifo of uart  */
extern struct cirbuf g_rx_fifo[UART_HW_NUM];

/** 
 * Initialisation function. This function puts the registers of the
 * microcontroler in a correct state in order to use the uart.  It
 * uses the configuration file <uart_config.h> ; this function is
 * equivalent to call uartX_setconf(NULL) for each uart.
 */
void uart_init(void);

/** 
 * Configure the uart 'num' with the given configuration. Returns 0 on
 * success.
 */
int8_t uart_setconf(uint8_t num, struct uart_config *u);

/** Get the current configuration of the uart 'num' */
void uart_getconf(uint8_t num, struct uart_config *u);

/** 
 * uart_recv returns the next character, taken from the fifo (if
 * any). If there is nothing to read, the function waits until
 * something come on the uart.
 */
int uart_recv(uint8_t num);

/** 
 * uart_recv returns the next character, taken from the fifo (if
 * any). If there is nothing to read, the function returns -1.
 */
int uart_recv_nowait(uint8_t num);

/**
 * same than uart_recv with 9 bits.
 */
int uart_9bits_recv(uint8_t num);

/** 
 * same than uart_recv_nowait() with 9 bits.
 */
int uart_9bits_recv_nowait(uint8_t num);

/**
 * uart_send_nowait is used to send data to the uart 'num'. The data
 * is first stored in the FIFO before beeing sent. If the FIFO is
 * full, data is dropped and the function returns -1, else it returns
 * the character c.
 */
int uart_send_nowait(uint8_t num, char c);

/**
 * uart_send is used to send data to the uart 'num'. The data is first
 * stored in the FIFO before beeing sent. If the FIFO is full, the
 * function wait until the uart is ready. The function returns c.
 */
int uart_send(uint8_t num, char c);

/**
 * uart_send_9bits is the same that uart_send but arg is 16 bits so
 * data can be 9 bits wide.
 */
int uart_send_9bits_nowait(uint8_t num, int c);

/* uart_send_9bits_wait is the same that uart_send_wait but arg is
 * 16 bits so data can be 9 bits wide.
 */
int uart_send_9bits(uint8_t num, int c);



/**
 * This function is used to register another function which will be
 * executed at each byte transmission (5, 6 ,7 ,8 bits)
 */
void uart_register_tx_event(uint8_t num, void (*f)(char));

/**
 * This function is used to register another function which will be
 * executed at each byte reception (5, 6 ,7 ,8 bits)
 */
void uart_register_rx_event(uint8_t num, void (*f)(char));


/**
 * This function is used to register another function which will be
 * executed at each 9 bits frame transmission. WARNING : it uses the
 * same internal pointer that the 8 bits event, so be carreful to 
 * unregister 8 bits events when doing 9 bits and vice versa.
 */
void uart_register_tx_9bits_event(uint8_t num, void (*f)(int));

/**
 * This function is used to register another function which will be
 * executed at each 9 bits reception. WARNING : it uses the
 * same internal pointer that the 8 bits event, so be carreful to 
 * unregister 8 bits events when doing 9 bits and vice versa.
 */
void uart_register_rx_9bits_event(uint8_t num, void (*f)(int));


/* global vars are initialized to 0 (NULL) */
event *rx_event[UART_HW_NUM];
event *tx_event[UART_HW_NUM];

SIGNAL(USART0_UDRE_vect)
{
	uart_send_next_char(0);
}

static void uart_recv_next_char(uint8_t num);

SIGNAL(USART0_RXC_vect)
{
	uart_recv_next_char(0);
}

/**
 * transmit next character of fifo if any, and call the event function.
 * This function is executed with intr locked.
 */
void uart_send_next_char(uint8_t num)
{
#ifdef CONFIG_MODULE_UART_9BITS
	if (uart_getconf_nbits(num) == 9) {
		uint16_t elt = 0;

		/* for 9 bits, it uses 2 places in the fifo */
		if (CIRBUF_GET_LEN(&g_tx_fifo[num]) < 2) {
			cbi(*uart_regs[num].ucsrb, UDRIE);
			return;
		}

		cirbuf_get_buf_tail(&g_tx_fifo[num], (char *)&elt, 2);
		cirbuf_del_buf_tail(&g_tx_fifo[num], 2);

		uart_set_udr_9bits(num, elt);
		sbi(*uart_regs[num].ucsrb, UDRIE);
	}
	else /* 5, 6, 7 or 8 bits */
#endif /* CONFIG_MODULE_UART_9BITS */
	{
		char elt = 0;

		if (CIRBUF_IS_EMPTY(&g_tx_fifo[num])) {
			cbi(*uart_regs[num].ucsrb, UDRIE);
			return;
		}

		elt = cirbuf_get_tail(&g_tx_fifo[num]);
		cirbuf_del_tail(&g_tx_fifo[num]);
		uart_set_udr(num, elt);
		sbi(*uart_regs[num].ucsrb, UDRIE);
	}
}

/**
 * UART RX Interrupt
 */
static void uart_recv_next_char(uint8_t num)
{
#ifdef CONFIG_MODULE_UART_9BITS
	if (uart_getconf_nbits() == 9) {
		int elt = 0;

		elt = uart_get_udr_9bits(num);
		if (CIRBUF_GET_FREELEN(&g_rx_fifo[num]) >= 2) {
			cirbuf_add_buf_head(&g_rx_fifo[num], (char *)&elt, 2);
		}

		if (rx_event[num])
			((event_9bits *)rx_event[num])(elt);
	}
	else
#endif /* CONFIG_MODULE_UART_9BITS */
	{
		char elt = 0;

		elt = uart_get_udr(num);
		if (!CIRBUF_IS_FULL(&g_rx_fifo[num])) {
			cirbuf_add_head(&g_rx_fifo[num], elt);
		}

		if (rx_event[num])
			rx_event[num](elt);
	}
}

/* This function is used to register another function which will be */
/* executed at each byte transmission. */
void uart_register_tx_event(uint8_t num, void (*f)(char))
{
	uint8_t flags;
	if (num >= UART_HW_NUM)
		return;
	IRQ_LOCK(flags);
	tx_event[num] = f;
	IRQ_UNLOCK(flags);
}

/* This function is used to register another function which will be */
/* executed at each byte reception */
void uart_register_rx_event(uint8_t num, void (*f)(char))
{
	uint8_t flags;
	if (num >= UART_HW_NUM)
		return;
	IRQ_LOCK(flags);
	rx_event[num] = f;
	IRQ_UNLOCK(flags);
}


/* get a char from the receive fifo */
int uart_9bits_recv_nowait(uint8_t num)
{
	char elt = 0;
	uint8_t flags;
	
	/* if interrupt mode is off, we have to check the status
	 * register */
	if (!(*uart_regs[num].ucsrb & (1 << RXCIE))) {
		if ( !(*uart_regs[num].ucsra & (1 << RXC)) )
			return -1;
		return uart_get_udr_9bits(num);
	}
	/* else check the fifo */
	else {
		IRQ_LOCK(flags);
		if( CIRBUF_GET_LEN(&g_rx_fifo[num]) >= 2) {
			cirbuf_get_buf_tail(&g_rx_fifo[num], (char *)&elt, 2);
			cirbuf_del_buf_tail(&g_rx_fifo[num], 2);
			IRQ_UNLOCK(flags);
			return (int)elt;
		}
		IRQ_UNLOCK(flags);
		return -1;
	}
}

/* get a char from the receive fifo */
int uart_9bits_recv(uint8_t num)
{
	int elt = 0;
	while ( (elt = uart_9bits_recv_nowait(num)) == -1 );
	return elt;
}

/* get a char from the receive fifo */
int uart_recv_nowait(uint8_t num)
{
	char elt = 0;
	uint8_t flags;
	
	/* if interrupt mode is off, we have to check the status
	 * register */
	if (!(*uart_regs[num].ucsrb & (1 << RXCIE))) {
		if ( !(*uart_regs[num].ucsra & (1 << RXC)) )
			return -1;
		return uart_get_udr(num);
	}
	/* else check the fifo */
	else {
		IRQ_LOCK(flags);
		if( !CIRBUF_IS_EMPTY(&g_rx_fifo[num]) ) {
			elt = cirbuf_get_tail(&g_rx_fifo[num]);
			cirbuf_del_tail(&g_rx_fifo[num]);
			IRQ_UNLOCK(flags);
			return (int)elt;
		}
		IRQ_UNLOCK(flags);
		return -1;
	}
}

/* get a char from the receive fifo */
int uart_recv(uint8_t num)
{
	int elt = 0;
	while ( (elt = uart_recv_nowait(num)) == -1 );
	return elt;
}

int uart_send_9bits_nowait(uint8_t num, int c)
{
	uint8_t flags;
	IRQ_LOCK(flags);

	/* if tx intrp are disabled (RXCIE is 0) */
	if ( !(*uart_regs[num].ucsrb & (1 << RXCIE )) ) {
		/* we have to poll the status register before xmit */
		if (*uart_regs[num].ucsra & (1<<UDRE)) {
			uart_set_udr_9bits(c);
			IRQ_UNLOCK(flags);
			return 0;
		}
		else {
			IRQ_UNLOCK(flags);
			return -1;
		}
	}

	/* the fifo must have 2 free places */
	if( CIRBUF_GET_FREELEN(&g_tx_fifo) < 2) {
		IRQ_UNLOCK(flags);
		return -1;
	}

	/* uart is ready to send */
	if (CIRBUF_IS_EMPTY(&g_tx_fifo[num]) && 
	    *uart_regs[num].ucsra & (1<<UDRE)) {
		uart_set_udr_9bits(c);
		sbi(*uart_regs[num].ucsrb, UDRIE); /* XXX */
	}
	else { /* not ready, put char in fifo */
		cirbuf_add_buf_head(&g_tx_fifo, (char *)&c, 2);
	}

	IRQ_UNLOCK(flags);
	return 0;
}

int uart_send_9bits(uint8_t num, int c)
{
	/* if cannot send the char */
	if (uart_send_9bits_nowait(num, c) == -1) {

		/* if irq lock are masked and interrupt mode is on, we
		 * have to poll the status register */
		if (GLOBAL_IRQ_ARE_MASKED() && (*uart_regs[num].ucsrb & (1 << RXCIE)) ) {
			while( !(*uart_regs[num].ucsra & (1 << UDRE)) );
			/* send the next char in the fifo to free two
			 * places */
			uart_send_next_char(num);
			cirbuf_add_buf_head(&g_tx_fifo[num], (char *)&c, 2);
		}
		else {
			/* if irq are not locked, we can loop to emit */
			while(uart_send_9bits_nowait(num, c) == -1);
		}
	}
	return 0;
}

int uart_send(uint8_t num, char c)
{
	/* if cannot send the char */
	if (uart_send_nowait(num, c) == -1) {

		/* if irq lock are masked and interrupt mode is on, we
		 * have to poll the status register */
		if (GLOBAL_IRQ_ARE_MASKED() && (*uart_regs[num].ucsrb & (1 << RXCIE)) ) {
			while( !(*uart_regs[num].ucsra & (1 << UDRE)) );
			/* send the next char in the fifo to free a
			 * place */
			uart_send_next_char(num);
			cirbuf_add_head(&g_tx_fifo[num], c);
		}
		else {
			/* if irq are not locked, we can loop to emit */
			while(uart_send_nowait(num, c) == -1);
		}
	}
	return 0;
}

/* send a char, or put it in the fifo if uart is not ready. Return -1
 * if fifo is full */
int uart_send_nowait(uint8_t num, char c)
{
	uint8_t flags;

	IRQ_LOCK(flags);

	/* if uart intrp mode is disabled (note that we look rx */
	/* intrp -- RXCIE is 0) */
	if ( !(*uart_regs[num].ucsrb & (1 << RXCIE )) ) {
		/* we have to poll the status register before xmit */
		if (*uart_regs[num].ucsra & (1<<UDRE)) {
			uart_set_udr(num, c);
			IRQ_UNLOCK(flags);
			return 0;
		}
		else {
			IRQ_UNLOCK(flags);
			return -1;
		}
	}

	if (CIRBUF_IS_FULL(&g_tx_fifo[num])) {
		IRQ_UNLOCK(flags);
		return -1;
	}

	/* uart is ready to send */
	if (CIRBUF_IS_EMPTY(&g_tx_fifo[num]) && 
	    *uart_regs[num].ucsra & (1<<UDRE)) {
		uart_set_udr(num, c);
		sbi(*uart_regs[num].ucsrb, UDRIE); /* XXX */
	}
	else { /* not ready, put char in fifo */
		cirbuf_add_head(&g_tx_fifo[num], c);
	}

	IRQ_UNLOCK(flags);
	return 0;
}

/* init all uarts */
void uart_init(void)
{
}

