/*
 * Copyright 2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <inttypes.h>
#include <unistd.h>
#include <string.h>

#include <stm32f4xx.h>
#include <gloss_device.h>

#include "uart.h"

//#define debug_printf(args...) printf(args)
#define debug_printf(args...)

struct spi_state {
	uint8_t cmd;             /* current command */
#define SPI_F_WREN 0x01          /* write enabled */
	uint8_t status;          /* status flags */
	uint32_t len;            /* number of rx/tx bytes since last reset */
	uint32_t addr;           /* current rd/wr address */
};

static struct spi_state spi_state;

#define SPI_DATA_LEN 8192        /* must be a power of 2 */
static uint8_t spi_data[SPI_DATA_LEN];

/* quick hack, approximately 1ms delay */
static void ms_delay(int ms)
{
	while (ms-- > 0) {
		volatile int x=5971;
		while (x-- > 0)
			__asm("nop");
	}
}

static void led_on(void)
{
	GPIOD->ODR |= (1 << 13);
}

static void led_off(void)
{
	GPIOD->ODR &= (~(1 << 13));
}

static int spi_init(void)
{
	SPI_InitTypeDef spi;
	GPIO_InitTypeDef gpio;

	/* Enable peripheral clock */
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_SPI1, ENABLE);

	/* Enable the AHB1 peripheral clock for GPIOA. */
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA, ENABLE);
	__asm("dsb");

	/* connect the pins to the desired alternate function,
	 * configure them: */
	GPIO_StructInit(&gpio);
	gpio.GPIO_Speed = GPIO_Speed_25MHz; /* common */
	gpio.GPIO_OType = GPIO_OType_PP;

	/* SPI1_NSS: PA4 */
	gpio.GPIO_Pin = GPIO_Pin_4;
	gpio.GPIO_Mode = GPIO_Mode_IN; /* do not use the hw NSS */
	GPIO_Init(GPIOA, &gpio);

	/* SPI1_SCK: PA5 */
	gpio.GPIO_Pin = GPIO_Pin_5;
	gpio.GPIO_Mode = GPIO_Mode_AF;
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource5, GPIO_AF_SPI1);
	GPIO_Init(GPIOA, &gpio);

	/* SPI1_MISO: PA6 */
	gpio.GPIO_Pin = GPIO_Pin_6;
	gpio.GPIO_Mode = GPIO_Mode_AF;
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource6, GPIO_AF_SPI1);
	GPIO_Init(GPIOA, &gpio);

	/* SPI1_MOSI: PA7 */
	gpio.GPIO_Pin = GPIO_Pin_7;
	gpio.GPIO_Mode = GPIO_Mode_AF;
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource7, GPIO_AF_SPI1);
	GPIO_Init(GPIOA, &gpio);

	/* Program the Polarity, Phase, First Data, Baud Rate Prescaler, Slave
	 * Management, Peripheral Mode and CRC Polynomial values */
	SPI_StructInit(&spi);
	spi.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
	spi.SPI_Mode = SPI_Mode_Slave;
	spi.SPI_DataSize = SPI_DataSize_8b;
	spi.SPI_CPOL = SPI_CPOL_Low;   /* CK to 0 when idle */
	spi.SPI_CPHA = SPI_CPHA_1Edge;  /* data on first transition */
	spi.SPI_NSS = SPI_NSS_Soft;    /* manage nss by software */
	spi.SPI_FirstBit = SPI_FirstBit_MSB;
	SPI_Init(SPI1, &spi);

	/* Enable the SPI */
	SPI_Cmd(SPI1, ENABLE);

	return 0;
}

static uint8_t spi_data_read(uint32_t addr)
{
	addr &= (SPI_DATA_LEN - 1);
	return spi_data[addr];
}

static void spi_data_write(uint32_t addr, uint8_t val)
{
	addr &= (SPI_DATA_LEN - 1);
	spi_data[addr] = val;
}

/* executed at startup */
static void spi_state_init(void)
{
	memset(&spi_state, 0, sizeof(spi_state));
}

/* executed between each command when nss goes low */
static void spi_state_reset(void)
{
	SPI_SendData(SPI1, 0);
	spi_state.cmd = 0;
	spi_state.len = 0;
	spi_state.addr = 0;
}

static void spi_data_intr(void)
{
	uint8_t rd_c;
	uint8_t wr_c = 0;

	rd_c = SPI_ReceiveData(SPI1);

	/* first byte, set command */
	if (spi_state.len == 0)
		spi_state.cmd = rd_c;

	switch(spi_state.cmd) {
	/* Read Memory, no dummy cycle */
	case 0x03:
		if (spi_state.len == 0) {
			spi_state.addr = 0;
		} else if (spi_state.len <= 3) {
			spi_state.addr <<= 8;
			spi_state.addr |= rd_c;
		}
		/* send data */
		if (spi_state.len >= 3) {
			wr_c = spi_data_read(spi_state.addr);
			spi_state.addr++;
		}
		break;
	/* Read Memory with dummy cycle */
	case 0x0B:
		if (spi_state.len == 0) {
			spi_state.addr = 0;
		} else if (spi_state.len <= 3) {
			spi_state.addr <<= 8;
			spi_state.addr |= rd_c;
		} else {
			wr_c = spi_data_read(spi_state.addr);
			spi_state.addr++;
		}
		break;

	/* Byte-Program (02H) */
	case 0x02:
		if (spi_state.len == 0) {
			spi_state.addr = 0;
		} else if (spi_state.len <= 3) {
			spi_state.addr <<= 8;
			spi_state.addr |= rd_c;
		} else if (spi_state.len == 4 &&
			(spi_state.status & SPI_F_WREN)) {
			spi_data_write(spi_state.addr, rd_c);
		} else {
			/* ignore next bytes */
		}
		break;

	/* Write-Enable (06H) */
	case 0x06:
		if (spi_state.len == 0)
			spi_state.status |= SPI_F_WREN;
		/* ignore next bytes */
		break;

	/* Write-Disable (04H) */
	case 0x04:
		if (spi_state.len == 0)
			spi_state.status &= (~SPI_F_WREN);
		/* ignore next bytes */
		break;

	/* Auto Address Increment Programming (ADH) */
	case 0xAD:
	/* Erase 4 KByte of memory array (20H) */
	case 0x20:
	/* Erase 32 KByte block of memory (52H) */
	case 0x52:
	/* Erase 64 KByte block of memory (D8H) */
	case 0xD8:
	/* Chip-Erase (60H) or (C7H) */
	case 0x60:
	case 0xC7:
	/* Read-Status-Register (05H) */
	case 0x05:
	/* Enable-Write-Status-Register (50H) */
	case 0x50:
	/* Write-Status-Register (01H) */
	case 0x01:
	/* Read-ID (90H) or (ABH) */
	case 0x90:
	case 0xAB:
	/* JEDEC-ID (9FH) */
	case 0x9F:
	/* EnableSOtooutputRY/BY# status during AAI programming (70H) */
	case 0x70:
	/* Disable SO as RY/BY# status during AAI programming (80H) */
	case 0x80:

	default:
		/* switch on LED on unknown command */
		led_on();
		break;
	}

	SPI_SendData(SPI1, wr_c);
	spi_state.len++;
	debug_printf("rx=0x%2.2x tx=0x%2.2x\n", rd_c, wr_c);
}

/* called on state transition on the NSS pin */
static void spi_nss_intr(uint8_t nss)
{
	if (nss) {
		debug_printf("ss high\n");
		SPI_NSSInternalSoftwareConfig(SPI1, SPI_NSSInternalSoft_Set);
	} else {
		debug_printf("ss low\n");
		SPI_NSSInternalSoftwareConfig(SPI1, SPI_NSSInternalSoft_Reset);
		/* it's a new command, reset our spi_state structure */
		spi_state_reset();
	}
}

int main(void)
{
	unsigned i;
	uint8_t prev_nss = 1, nss;

	/* enable the clock to GPIOD, and stall instruction pipeline as per
	 * errata 2.1.13 "Delay after an RCC peripheral clock enabling" */
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOD, ENABLE);
	__asm("dsb");

	/* set pin 13 to be general purpose output */
	GPIOD->MODER = (1 << 26);

	/* toggle the pin */
	for (i = 0; i < 3; i++) {
		led_on();
		ms_delay(500);
		led_off();
		ms_delay(500);
	}

	uart_init();
	uart_register_stdio();
	printf("salut\n");

	if (spi_init() < 0) {
		printf("SPI init failed\n");
		goto end;
	}

	spi_state_init();

	/* init spi data */
	memset(spi_data, 0, sizeof(spi_data));
	strcpy((char *)&spi_data[0x1000], "Dr0idZ C0rp");

	while (1) {
		/* inspect nss changes */
		nss = !!(GPIOA->IDR & (1 << 4));
		if (nss != prev_nss) {
			spi_nss_intr(nss);
			prev_nss = nss;
		}

		/* data received */
		if (SPI_GetFlagStatus(SPI1, SPI_FLAG_RXNE) != RESET)
			spi_data_intr();
	}

 end:
	printf("end\n");
	while (1);
	return 0;
}
