/*
 * Copyright 2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <inttypes.h>
#include <unistd.h>
#include "stm32f4xx.h"

#include "gloss_device.h"

static _ssize_t
uart_write_r(__attribute__((unused)) struct _reent *r,
	__attribute__((unused)) int fd,
	const void *ptr, size_t len)
{
	size_t i;
	const char *buf = ptr;


	for (i = 0; i < len; i++) {
		/* wait that uart is ready */
		while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET)
			;
		USART_SendData(USART2, buf[i]);
	}
	return len;
}

static _ssize_t
uart_read_r(__attribute__((unused)) struct _reent *r,
	__attribute__((unused)) int fd,
	void *ptr, size_t len)
{
	size_t i;
	char *buf = ptr;

	for (i = 0; i < len; i++) {
		/* wait that uart is ready */
		while (USART_GetFlagStatus(USART2, USART_FLAG_RXNE) == RESET)
			;
		buf[i] = USART_ReceiveData(USART2);
	}
	return len;
}

struct libc_device uart_stdin_dev = {
	.name = "stdin",
	.open_r = NULL,
	.close_r = NULL,
	.read_r = uart_read_r,
	.write_r = NULL,
};

struct libc_device uart_stdout_dev = {
	.name = "stdout",
	.open_r = NULL,
	.close_r = NULL,
	.read_r = NULL,
	.write_r = uart_write_r,
};

struct libc_device uart_stderr_dev = {
	.name = "stderr",
	.open_r = NULL,
	.close_r = NULL,
	.read_r = NULL,
	.write_r = uart_write_r,
};

void uart_init(void)
{
	USART_InitTypeDef uart;
	GPIO_InitTypeDef gpio;

	/* Enable the USART2 peripheral clock. */
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2, ENABLE);
	__asm("dsb");

	/* Enable the AHB1 peripheral clock for GPIOA. */
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA, ENABLE);
	__asm("dsb");

	/* connect to alternate function */
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource2, GPIO_AF_USART2);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource3, GPIO_AF_USART2);

	GPIO_StructInit(&gpio);
	gpio.GPIO_Speed = GPIO_Speed_25MHz;

	/* configure tx on PA2 */
	gpio.GPIO_Pin = GPIO_Pin_2;
	gpio.GPIO_Mode = GPIO_Mode_AF;
	GPIO_Init(GPIOA, &gpio);

	/* configure rx on PA3 */
	gpio.GPIO_Pin = GPIO_Pin_3;
	gpio.GPIO_Mode = GPIO_Mode_AF;
	GPIO_Init(GPIOA, &gpio);

	/* USART configuration */
	uart.USART_BaudRate = 115200;
	uart.USART_WordLength = USART_WordLength_8b;
	uart.USART_StopBits = USART_StopBits_1;
	uart.USART_Parity = USART_Parity_No;
	uart.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
	uart.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
	USART_Init(USART2, &uart);

	/* Enable USART */
	USART_Cmd(USART2, ENABLE);
}

void uart_register_stdio(void)
{
	/* register device twice (stdin then stdout) */
	libc_device_register(&uart_stdin_dev);
	libc_device_register(&uart_stdout_dev);
	libc_device_register(&uart_stderr_dev);

	/* Disable buffering */
	setbuf(stdin, NULL);
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);
}

