/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <errno.h>

#include "cirbuf.h"

/* init a circular buffer */
void
cirbuf_init(struct cirbuf *cbuf, char *buf, unsigned int start,
	unsigned int maxlen)
{
	cbuf->maxlen = maxlen;
	cbuf->len = 0;
	cbuf->start = start;
	cbuf->end = start;
	cbuf->buf = buf;
}

/* multiple add at head */
int
cirbuf_add_buf_head(struct cirbuf *cbuf, const char *c, unsigned int n)
{
	unsigned int e;

	if (!n || n > CIRBUF_GET_FREELEN(cbuf))
		return -EINVAL;

	e = CIRBUF_IS_EMPTY(cbuf) ? 1 : 0;

	if (n < cbuf->start + e) {
		memcpy(cbuf->buf + cbuf->start - n + e, c, n);
	}
	else {
		memcpy(cbuf->buf, c  + n - (cbuf->start + e) , cbuf->start + e);
		memcpy(cbuf->buf + cbuf->maxlen - n + (cbuf->start + e), c,
		       n - (cbuf->start + e));
	}
	cbuf->len += n;
	cbuf->start += (cbuf->maxlen - n + e);
	cbuf->start %= cbuf->maxlen;
	return n;
}

/* multiple add at tail */
int
cirbuf_add_buf_tail(struct cirbuf *cbuf, const char *c, unsigned int n)
{
	unsigned int e;

	if (!n || n > CIRBUF_GET_FREELEN(cbuf))
		return -EINVAL;

	e = CIRBUF_IS_EMPTY(cbuf) ? 1 : 0;

	if (n < cbuf->maxlen - cbuf->end - 1 + e) {
		memcpy(cbuf->buf + cbuf->end + !e, c, n);
	}
	else {
		memcpy(cbuf->buf + cbuf->end + !e, c, cbuf->maxlen -
		       cbuf->end - 1 + e);
		memcpy(cbuf->buf, c + cbuf->maxlen - cbuf->end - 1 + e,
		       n - cbuf->maxlen + cbuf->end + 1 - e);
	}
	cbuf->len += n;
	cbuf->end += n - e;
	cbuf->end %= cbuf->maxlen;
	return n;
}

/* single add at head */
static inline void
__cirbuf_add_head(struct cirbuf * cbuf, char c)
{
	if (!CIRBUF_IS_EMPTY(cbuf)) {
		cbuf->start += (cbuf->maxlen - 1);
		cbuf->start %= cbuf->maxlen;
	}
	cbuf->buf[cbuf->start] = c;
	cbuf->len ++;
}

/* single add at head, checking if full first */
int
cirbuf_add_head_safe(struct cirbuf * cbuf, char c)
{
	if (cbuf && !CIRBUF_IS_FULL(cbuf)) {
		__cirbuf_add_head(cbuf, c);
		return 0;
	}
	return -EINVAL;
}

/* single add at head */
void
cirbuf_add_head(struct cirbuf * cbuf, char c)
{
	__cirbuf_add_head(cbuf, c);
}

/* single add at tail */
static inline void
__cirbuf_add_tail(struct cirbuf * cbuf, char c)
{
	if (!CIRBUF_IS_EMPTY(cbuf)) {
		cbuf->end ++;
		cbuf->end %= cbuf->maxlen;
	}
	cbuf->buf[cbuf->end] = c;
	cbuf->len ++;
}

/* single add at tail, checking if full first */
int
cirbuf_add_tail_safe(struct cirbuf * cbuf, char c)
{
	if (cbuf && !CIRBUF_IS_FULL(cbuf)) {
		__cirbuf_add_tail(cbuf, c);
		return 0;
	}
	return -EINVAL;
}

/* single add at tail */
void
cirbuf_add_tail(struct cirbuf * cbuf, char c)
{
	__cirbuf_add_tail(cbuf, c);
}

/* shift the buffer */
static inline void
__cirbuf_shift_left(struct cirbuf *cbuf)
{
	unsigned int i;
	char tmp = cbuf->buf[cbuf->start];

	for (i=0 ; i<cbuf->len ; i++) {
		cbuf->buf[(cbuf->start+i)%cbuf->maxlen] =
			cbuf->buf[(cbuf->start+i+1)%cbuf->maxlen];
	}
	cbuf->buf[(cbuf->start-1+cbuf->maxlen)%cbuf->maxlen] = tmp;
	cbuf->start += (cbuf->maxlen - 1);
	cbuf->start %= cbuf->maxlen;
	cbuf->end += (cbuf->maxlen - 1);
	cbuf->end %= cbuf->maxlen;
}

static inline void
__cirbuf_shift_right(struct cirbuf *cbuf)
{
	unsigned int i;
	char tmp = cbuf->buf[cbuf->end];

	for (i=0 ; i<cbuf->len ; i++) {
		cbuf->buf[(cbuf->end+cbuf->maxlen-i)%cbuf->maxlen] =
			cbuf->buf[(cbuf->end+cbuf->maxlen-i-1)%cbuf->maxlen];
	}
	cbuf->buf[(cbuf->end+1)%cbuf->maxlen] = tmp;
	cbuf->start += 1;
	cbuf->start %= cbuf->maxlen;
	cbuf->end += 1;
	cbuf->end %= cbuf->maxlen;
}

/* XXX we could do a better algorithm here... */
void cirbuf_align_left(struct cirbuf * cbuf)
{
	if (cbuf->start < cbuf->maxlen/2) {
		while (cbuf->start != 0) {
			__cirbuf_shift_left(cbuf);
		}
	}
	else {
		while (cbuf->start != 0) {
			__cirbuf_shift_right(cbuf);
		}
	}
}

/* XXX we could do a better algorithm here... */
void cirbuf_align_right(struct cirbuf * cbuf)
{
	if (cbuf->start >= cbuf->maxlen/2) {
		while (cbuf->end != cbuf->maxlen-1) {
			__cirbuf_shift_left(cbuf);
		}
	}
	else {
		while (cbuf->start != cbuf->maxlen-1) {
			__cirbuf_shift_right(cbuf);
		}
	}
}

/* multiple delete at head */
int
cirbuf_del_buf_head(struct cirbuf *cbuf, unsigned int size)
{
	if (!size || size > CIRBUF_GET_LEN(cbuf))
		return -EINVAL;

	cbuf->len -= size;
	if (CIRBUF_IS_EMPTY(cbuf)) {
		cbuf->start += size - 1;
		cbuf->start %= cbuf->maxlen;
	}
	else {
		cbuf->start += size;
		cbuf->start %= cbuf->maxlen;
	}
	return 0;
}

/* multiple delete at tail */
int
cirbuf_del_buf_tail(struct cirbuf *cbuf, unsigned int size)
{
	if (!size || size > CIRBUF_GET_LEN(cbuf))
		return -EINVAL;

	cbuf->len -= size;
	if (CIRBUF_IS_EMPTY(cbuf)) {
		cbuf->end  += (cbuf->maxlen - size + 1);
		cbuf->end %= cbuf->maxlen;
	}
	else {
		cbuf->end  += (cbuf->maxlen - size);
		cbuf->end %= cbuf->maxlen;
	}
	return 0;
}

/* single del at head */
static inline void
__cirbuf_del_head(struct cirbuf * cbuf)
{
	cbuf->len --;
	if (!CIRBUF_IS_EMPTY(cbuf)) {
		cbuf->start ++;
		cbuf->start %= cbuf->maxlen;
	}
}

/* single del at head, checking if empty first */
int
cirbuf_del_head_safe(struct cirbuf * cbuf)
{
	if (cbuf && !CIRBUF_IS_EMPTY(cbuf)) {
		__cirbuf_del_head(cbuf);
		return 0;
	}
	return -EINVAL;
}

/* single del at head */
void
cirbuf_del_head(struct cirbuf * cbuf)
{
	__cirbuf_del_head(cbuf);
}

/* single del at tail */
static inline void
__cirbuf_del_tail(struct cirbuf * cbuf)
{
	cbuf->len --;
	if (!CIRBUF_IS_EMPTY(cbuf)) {
		cbuf->end  += (cbuf->maxlen - 1);
		cbuf->end %= cbuf->maxlen;
	}
}

/* single del at tail, checking if empty first */
int
cirbuf_del_tail_safe(struct cirbuf * cbuf)
{
	if (cbuf && !CIRBUF_IS_EMPTY(cbuf)) {
		__cirbuf_del_tail(cbuf);
		return 0;
	}
	return -EINVAL;
}

/* single del at tail */
void
cirbuf_del_tail(struct cirbuf * cbuf)
{
	__cirbuf_del_tail(cbuf);
}

/* convert to buffer */
int
cirbuf_get_buf_head(struct cirbuf *cbuf, char *c, unsigned int size)
{
	unsigned int n;

	n = (size < CIRBUF_GET_LEN(cbuf)) ? size : CIRBUF_GET_LEN(cbuf);

	if (!n)
		return 0;

	if (cbuf->start <= cbuf->end) {
		memcpy(c, cbuf->buf + cbuf->start , n);
	}
	else {
		memcpy(c, cbuf->buf + cbuf->start , cbuf->maxlen - cbuf->start);
		memcpy(c + cbuf->maxlen - cbuf->start, cbuf->buf,
		       n - cbuf->maxlen + cbuf->start);
	}
	return n;
}

/* convert to buffer */
int
cirbuf_get_buf_tail(struct cirbuf *cbuf, char *c, unsigned int size)
{
	unsigned int n;

	n = (size < CIRBUF_GET_LEN(cbuf)) ? size : CIRBUF_GET_LEN(cbuf);

	if (!n)
		return 0;

	if (cbuf->start <= cbuf->end) {
		memcpy(c, cbuf->buf + cbuf->end - n + 1, n);
	}
	else {
		memcpy(c + cbuf->maxlen - cbuf->start,
		       cbuf->buf, cbuf->end + 1);
		memcpy(c, cbuf->buf + cbuf->maxlen - n + cbuf->end +1,
		       n - cbuf->end - 1);
	}
	return n;
}

/* get head */
char
cirbuf_get_head(struct cirbuf * cbuf)
{
	return cbuf->buf[cbuf->start];
}

/* get tail */
char
cirbuf_get_tail(struct cirbuf * cbuf)
{
	return cbuf->buf[cbuf->end];
}

