/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <inttypes.h>
#include <fcntl.h>

#include "cmdline_parse.h"
#include "cmdline_rdline.h"
#include "cmdline.h"

static void
cmdline_default_valid_buffer(struct rdline *rdl, const char *buf,
			     __attribute__((unused)) unsigned int size)
{
	struct cmdline *cl = rdl->opaque;
	int ret;
	ret = cmdline_parse(cl->ctx, buf, cl);
	if (ret == CMDLINE_PARSE_AMBIGUOUS)
		cmdline_printf(cl, "Ambiguous command\n");
	else if (ret == CMDLINE_PARSE_NOMATCH)
		cmdline_printf(cl, "Bad arguments\n");
	else if (ret == CMDLINE_PARSE_UNTERMINATED_QUOTE)
		cmdline_printf(cl, "Unterminated quote\n");
}

static int
cmdline_default_complete_buffer(struct rdline *rdl, const char *buf,
				char *dstbuf, unsigned int dstsize)
{
	struct cmdline *cl = rdl->opaque;
	// RDLINE_RES_COMPLETE ?
	return cmdline_complete(cl->ctx, buf, dstbuf, dstsize);
}


static int
cmdline_default_help(struct rdline *rdl, const char *buf,
		     rdline_write_t *write_cb, void *write_arg)
{
	struct cmdline *cl = rdl->opaque;
	return cmdline_help(cl->ctx, buf, cmdline_write, write_arg);
}

/* ---- Some rdline wrappers ---- */

/* write a buffer, just use rdline */
ssize_t
cmdline_write(void *arg, void *buf, size_t count)
{
	struct rdline *rdl = arg;
#ifdef NO_PAGER
	return rdline_write(rdl, buf, count);
#else
	return rdline_asyncpager_write(rdl, buf, count);
#endif
}

void
cmdline_set_prompt(struct cmdline *cl, const char *prompt)
{
	snprintf(cl->prompt, sizeof(cl->prompt), "%s", prompt);
}

void
cmdline_init(struct cmdline *cl, cmdline_parse_ctx_t *ctx,
	const char *prompt, int s_in, int s_out)
{
	/* init cmdline structure */
	memset(cl, 0, sizeof(struct cmdline));
	cl->s_in = s_in;
	cl->s_out = s_out;
	cl->ctx = ctx;

	/* init embedded rdline */
	rdline_init(&cl->rdl, s_in, s_out,
		    cmdline_default_valid_buffer,
		    cmdline_default_complete_buffer,
		    cmdline_default_help);

	cl->rdl.opaque = cl;
	cmdline_set_prompt(cl, prompt);
	rdline_newline(&cl->rdl, cl->prompt);
}

struct cmdline *
cmdline_new(cmdline_parse_ctx_t *ctx, const char *prompt, int s_in, int s_out)
{
	struct cmdline *cl;

	cl = malloc(sizeof(struct cmdline));
	if (cl == NULL)
		return NULL;

	cmdline_init(cl, ctx, prompt, s_in, s_out);
	return cl;
}

void
cmdline_free(struct cmdline *cl)
{
	if (cl->s_in > 2)
		close(cl->s_in);
	if (cl->s_out != cl->s_in && cl->s_out > 2)
		close(cl->s_out);
	free(cl);
}

int
cmdline_printf(const struct cmdline *cl, const char *fmt, ...)
{
	va_list ap;
	int ret;

	va_start(ap, fmt);
	ret = rdline_vprintf(&cl->rdl, fmt, ap);
	va_end(ap);

	return ret;
}

/* Push an input buffer in the command line. Typically, this function
 * is called by cmdline_interact() to send the input characters to the
 * cmdline process. It can also be called by a user callback function,
 * when a buffer is received from the input socket.
 *
 * The function returns the number of processed characters, or a
 * negative value on error (EOF reached or command line exited. */
int
cmdline_in(struct cmdline *cl, const char *buf, int size)
{
	int ret = 0;
	int i;

	for (i = 0; i < size; i++) {
		ret = rdline_char_in(&cl->rdl, buf[i]);

		if (ret == RDLINE_RES_VALIDATED &&
		    cl->rdl.status == RDLINE_STOPPED)
			break;

		if (ret == RDLINE_RES_VALIDATED)
			rdline_newline(&cl->rdl, cl->prompt);
		else if (ret == RDLINE_RES_EOF)
			return -1;
		else if (ret == RDLINE_RES_EXITED)
			return -1;
	}
	return i;
}

/* Interrupt a running command line process */
void
cmdline_quit(struct cmdline *cl)
{
	rdline_quit(&cl->rdl);
}

/* Start command line on configured file descriptor. This function
 * loops until the user explicitelly call cmdline_quit(), or if the
 * input fd reaches EOF. */
void
cmdline_interact(struct cmdline *cl)
{
	char c;

	c = -1;
	while (1) {
		if (read(cl->s_in, &c, 1) <= 0)
			break;
		if (cmdline_in(cl, &c, 1) < 0)
			break;
	}
}
