/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>

#include "cmdline_parse.h"
#include "cmdline.h"

//#define debug_printf printf
#define debug_printf(args...) do {} while(0)

/* used internally for cmdline_help() and cmdline_complete() */
struct cmdline_preparse {
	int nb_valid_tok; /* number of valid tokens in the buffer */
	void *opaque;     /* pointer to opaque data */
	char comp_tok_buf[CMDLINE_MAX_TOKEN_SIZE]; /* token to complete */
	int comp_tok_len; /* length of the token to complete */
	int comp_tok_offset; /* offset of token to complete in the line buf */
};

/* isblank() needs _XOPEN_SOURCE >= 600 || _ISOC99_SOURCE, so use our
 * own. */
static int
isblank2(char c)
{
	if (c == ' ' ||
	    c == '\t' )
		return 1;
	return 0;
}

static int
isendofline(char c)
{
	if (c == '\n' ||
	    c == '\r' )
		return 1;
	return 0;
}

static int
iscomment(char c)
{
	if (c == '#')
		return 1;
	return 0;
}

int
cmdline_isendoftoken(char c)
{
	if (!c || iscomment(c) || isblank2(c) || isendofline(c))
		return 1;
	return 0;
}

static unsigned int
nb_common_chars(const char * s1, const char * s2)
{
	unsigned int i=0;

	while (*s1==*s2 && *s1 && *s2) {
		s1++;
		s2++;
		i++;
	}
	return i;
}

/* quote a string and escape original quotes */
int cmdline_quote_token(char *dst, unsigned dstlen, const char *src)
{
	unsigned s = 0, d = 0;

	/* the 2 quotes + '\0' */
	if (dstlen < 3)
		return -EMSGSIZE;

	dst[d++] = '"';
	while (src[s] != '\0') {
		if (d >= (dstlen-2))
			return -EMSGSIZE;

		if (src[s] == '"')
			dst[d++] = '\\';
		if (src[s] == '\\' && src[s+1] == '"')
			dst[d++] = '\\';

		dst[d++] = src[s++];
	}

	if (d >= (dstlen-2))
		return -EMSGSIZE;
	dst[d++] = '"';
	dst[d++] = '\0';
	return s;
}

/* Remove quote and stop when we reach the end of token. Return the
 * number of "eaten" bytes from the source buffer, or a negative value
 * on error */
int cmdline_get_token(char *dst, unsigned dstlen, const char *src)
{
	unsigned s = 0, d = 0;
	int quoted = 0;

	/* skip spaces */
	while (isblank2(src[s]))
		s++;

	/* empty token */
	if (cmdline_isendoftoken(src[s]))
		return -EINVAL;

	/* copy token and remove quotes */
	while (src[s] != '\0') {
		if (d >= dstlen)
			return -EMSGSIZE;

		if (cmdline_isendoftoken(src[s]) && quoted == 0)
			break;

		if (src[s] == '\\' && src[s+1] == '"') {
			dst[d++] = '"';
			s += 2;
			continue;
		}
		if (src[s] == '\\' && src[s+1] == '\\') {
			dst[d++] = '\\';
			s += 2;
			continue;
		}
		if (src[s] == '"') {
			s++;
			quoted = !quoted;
			continue;
		}
		dst[d++] = src[s++];
	}

	/* not enough room in dst buffer */
	if (d >= (dstlen-1))
		return -EMSGSIZE;

	/* end of string during quote */
	if (quoted)
		return -EINVAL;

	dst[d++] = '\0';
	return  s;
}

/* return the nth token from src and copy it in dst. Return the offset
 * of the token in src, or a negative value on error. Note: the index
 * of the first token is 0. */
static int cmdline_get_nth_token(char *dst, unsigned dstlen, int n,
				 const char *src)
{
	int ret = 0, offset = 0;

	do {
		offset += ret;

		/* skip spaces */
		while (isblank2(src[offset]))
			offset++;

		/* get the token starting at offset */
		ret = cmdline_get_token(dst, dstlen, src + offset);
		if (ret < 0)
			return ret;

	} while (n--);

	return offset;
}

/*
 * try to match the buffer with an instruction (only the first
 * nb_match_token tokens if != 0). Return 0 if we match all the
 * tokens, else the number of matched tokens, else -1.

retval à vérifier.

 */
static int
match_inst(const cmdline_parse_inst_t *inst, const char *linebuf,
	   unsigned int nb_match_token, void *resbuf, unsigned resbuf_size)
{
	unsigned int token_num = 0;
	cmdline_parse_token_hdr_t *token;
	int n = 0, res;
	char token_str[CMDLINE_MAX_TOKEN_SIZE];

	token = inst->tokens[token_num];

	/* check if we match all tokens of inst */
	while (token) {

		/* we matched enough tokens, return success */
		if (nb_match_token != 0 && token_num >= nb_match_token)
			return 0;

		debug_printf("TK\n");

		/* copy token and remove quotes */
		n = cmdline_get_token(token_str, sizeof(token_str), linebuf);
		if (n < 0)
			break;

		/* parse this token */
		if (resbuf == NULL)
			res = token->ops->parse(token, token_str, NULL, 0);
		else {
			unsigned rb_sz;
			void *rb = (char *)resbuf + token->offset;

			/* not enough room to store result */
			if (token->offset > resbuf_size)
				return -ENOBUFS;

			rb_sz = resbuf_size - token->offset;
			res = token->ops->parse(token, token_str, rb, rb_sz);
		}

		/* does not match this token */
		if (res < 0)
			break;

		debug_printf("TK parsed (len=%d)\n", n);
		linebuf += n;
		token_num ++;
		token = inst->tokens[token_num];
	}

	/* does not match */
	if (token_num == 0)
		return -ENOENT;

	/* we don't match all the tokens */
	if (token)
		return token_num;

	/* are there are some tokens more */
	while (isblank2(*linebuf))
		linebuf++;

	/* end of buf, we match all inst  */
	if (*linebuf == '\0' || isendofline(*linebuf) || iscomment(*linebuf))
		return 0;

	/* garbage after inst */
	return token_num;
}


/* Check if a line buffer is valid and can be parsed or completed. The
 * parsing stops when \n or \0 is reached. The also function checks
 * that tokens are correctly quoted. The number of tokens in the
 * buffer is returned. */
static int cmdline_validate_linebuf(const char *linebuf)
{
	int quoted = 0, comment = 0;
	int i = 0, nbtok = 0, token = 0;

	while (linebuf[i] != '\0') {
		if (isendofline(linebuf[i]) && quoted == 0)
			break;
		if (comment == 1) {
			i++;
			continue;
		}
		if (iscomment(linebuf[i]) && quoted == 0) {
			comment = 1;
			i ++;
			continue;
		}

		/* end of token */
		if (isblank2(linebuf[i]) && quoted == 0)
			token = 0;
		/* new token */
		if (!isblank2(linebuf[i]) && token == 0) {
			token = 1;
			nbtok++;
		}

		if (linebuf[i] == '\\' && linebuf[i+1] == '"') {
			i += 2;
			continue;
		}
		if (linebuf[i] == '\\' && linebuf[i+1] == '\\') {
			i += 2;
			continue;
		}
		if (linebuf[i] == '"') {
			i++;
			quoted = !quoted;
			continue;
		}
		i++;
	}
	if (quoted)
		return CMDLINE_PARSE_UNTERMINATED_QUOTE;
	return nbtok;
}

/* Try to parse a buffer according to the specified context. The
 * argument linebuf must end with \n or \0. */
int
cmdline_parse(cmdline_parse_ctx_t *ctx, const char *linebuf, void *opaque)
{
	const cmdline_parse_inst_t **pinst;
	const cmdline_parse_inst_t *inst;
	char result_buf[CMDLINE_MAX_DSTBUF_SIZE];
	void (*f)(void *, struct cmdline *, void *) = NULL;
	void *data = NULL;
	int ret;

	ret = cmdline_validate_linebuf(linebuf);
	if (ret < 0)
		return ret;
	if (ret == 0)
		return CMDLINE_PARSE_EMPTY;


	/* parse it !! */
	for (pinst = &ctx->insts[0]; *pinst != NULL; pinst++) {
		inst = *pinst;
		debug_printf("INST\n");

		/* fully parsed */
		ret = match_inst(inst, linebuf, 0, result_buf, sizeof(result_buf));

		if (ret != 0)
			continue;

		debug_printf("INST fully parsed\n");

		/* if end of buf -> there is no garbage after inst */
		if (f != NULL) {
			/* more than 1 inst matches */
			debug_printf("Ambiguous cmd\n");
			return CMDLINE_PARSE_AMBIGUOUS;
		}
		f = inst->f;
		data = inst->data;
	}

	/* call func */
	if (f == NULL)
		return CMDLINE_PARSE_NOMATCH;

	f(result_buf, opaque, data);
	return CMDLINE_PARSE_SUCCESS;
}

/* called by cmdline_help() and cmdline_complete() to preparse the
 * line buffer (the operations done are common to these functions) */
static int cmdline_preparse(struct cmdline_preparse *preparse, const char *buf)
{
	int ret, len, nb_tok;

	/* count the number of tokens in the line buffer */
	ret = cmdline_validate_linebuf(buf);
	if (ret < 0)
		return ret;
	nb_tok = ret;

	/* if last token is not complete, decrement nb_valid_tok */
	len = strlen(buf);
	if (nb_tok == 0 || isblank2(buf[len - 1])) {
		preparse->nb_valid_tok = nb_tok;
		preparse->comp_tok_offset = len;
		preparse->comp_tok_buf[0] = '\0';
		preparse->comp_tok_len = 0;
	}
	else {
		preparse->nb_valid_tok = nb_tok - 1;

		/* get the incomplete token (can be empty) and return its
		 * offset in the buffer */
		preparse->comp_tok_offset =
			cmdline_get_nth_token(preparse->comp_tok_buf,
					      sizeof(preparse->comp_tok_buf),
					      preparse->nb_valid_tok,
					      buf);
		preparse->comp_tok_len = strlen(preparse->comp_tok_buf);
	}

	return 0;
}

/* Display a contextual help using the write_buf() function pointer
 * given as parameter (called with its opaque pointer). The contextual
 * help depends on the buffer given. */
int cmdline_help(cmdline_parse_ctx_t *ctx, const char *buf,
		 cmdline_write_t *write_buf, void *opaque)
{
	cmdline_parse_token_hdr_t *token;
	const cmdline_parse_inst_t **pinst;
	const cmdline_parse_inst_t *inst;
	struct cmdline_preparse preparse;
	char helpbuf[CMDLINE_MAX_DSTBUF_SIZE];
	char helpbuf2[CMDLINE_MAX_DSTBUF_SIZE];
	char tmpbuf[CMDLINE_MAX_DSTBUF_SIZE];
	char *help_str;
	int n, m, iterate;

	cmdline_preparse(&preparse, buf);

	debug_printf("display contextual help\n");

	for (pinst = &ctx->insts[0]; *pinst != NULL; pinst++) {
		inst = *pinst;

		/* match the beginning of the command */
		if (preparse.nb_valid_tok != 0 &&
		    match_inst(inst, buf, preparse.nb_valid_tok,
			       NULL, 0) != 0)
			continue;

		token = inst->tokens[preparse.nb_valid_tok];

		/* end of inst */
		if (token == NULL) {
			n = snprintf(helpbuf, sizeof(helpbuf), "[Return]\n");
			if (n > 0)
				write_buf(opaque, helpbuf, n);
			continue;
		}

		/* token matches, but no completion */
		if (token->ops->complete_start == NULL ||
		    token->ops->complete_iterate == NULL)
			iterate = 0;
		else
			iterate = 1;

		/* store the incomplete token in tmpbuf */
		n = preparse.comp_tok_len + 1;
		if (n > sizeof(tmpbuf))
			n = sizeof(tmpbuf);
		snprintf(tmpbuf, n, "%s", preparse.comp_tok_buf);

		if (iterate == 1 &&
		    token->ops->complete_start(token, tmpbuf,
					       &preparse.opaque) < 0) {
			/* cancel iteration, complete_start() returned
			 * a negative value, meaning no completion  */
			iterate = 0;
			if (token->ops->complete_end != NULL)
				token->ops->complete_end(token,
							 &preparse.opaque);
		}

		/* get token dynamic help string */
		if ((token->ops->help == NULL) ||
		    (token->ops->help(token, tmpbuf, sizeof(tmpbuf)) < 0))
			snprintf(tmpbuf, sizeof(tmpbuf), "unknown");

		/* get instruction static help string */
		help_str = inst->help_str;
		if (help_str == NULL)
			help_str = "No help";

		/* get the help for this token */
		n = snprintf(helpbuf, sizeof(helpbuf), "[%s]: %s\n",
			     tmpbuf, help_str);

		if (iterate == 0)
			continue;

		/* iterate over all possible completion for this inst */
		while (token->ops->complete_iterate(token,
						    &preparse.opaque,
						    tmpbuf,
						    sizeof(tmpbuf)) >= 0) {


			debug_printf("   choice <%s>\n", tmpbuf);

			/* does the completion match the beginning of
			 * the word ? */
			if (strncmp(preparse.comp_tok_buf, tmpbuf,
				    preparse.comp_tok_len))
				continue;

			/* get the token and add it in help buffer */
			m = snprintf(helpbuf2, sizeof(helpbuf2), "%-20s ",
				     tmpbuf);
			if (m > 0)
				write_buf(opaque, helpbuf2, m);

			/* display quotes if help was already displayed */
			if (n == 0) {
				for (n = 0; m < 60; n++, m++)
					helpbuf[n] = ' ';
				helpbuf[n++] = '"';
				helpbuf[n++] = '\n';
				helpbuf[n] = '\0';
			}
			if (n > 0)
				write_buf(opaque, helpbuf, n);

			/* don't display help next time */
			n = 0;
		}

		/* no more completion, go to next inst */
		if ( token->ops->complete_end != NULL)
			token->ops->complete_end(token, &preparse.opaque);
	}

	return 0;
}

/* try to complete the buffer given as a parameter */
int
cmdline_complete(cmdline_parse_ctx_t *ctx, const char *buf,
		 char *dst, unsigned int dstsize)
{
	cmdline_parse_token_hdr_t *token;
	const cmdline_parse_inst_t **pinst;
	const cmdline_parse_inst_t *inst;
	struct cmdline_preparse preparse;
	int nb_match = 0;
	int nb_completion = 0;
	int completion_len = CMDLINE_MAX_TOKEN_SIZE;
	char completion_buf[CMDLINE_MAX_TOKEN_SIZE];
	char tmpbuf[CMDLINE_MAX_TOKEN_SIZE];
	int ret, n;

	debug_printf("%s called\n", __FUNCTION__);

	/* fill the preparse structure that contains infos that will
	 * help us to complete the buffer */
	ret = cmdline_preparse(&preparse, buf);
	if (ret < 0)
		return CMDLINE_COMPLETE_NONE;

	/* try to complete !! */
	for (pinst = &ctx->insts[0]; *pinst != NULL; pinst++) {
		inst = *pinst;

		debug_printf("INST\n");

		/* try to match the first tokens */
		if (preparse.nb_valid_tok != 0 &&
		    match_inst(inst, buf, preparse.nb_valid_tok,
			       NULL, 0) != 0)
			continue;

		nb_match ++;
		token = inst->tokens[preparse.nb_valid_tok];

		/* non completable */
		if (token == NULL ||
		    token->ops->complete_start == NULL ||
		    token->ops->complete_iterate == NULL)
			continue;

		/* store the incomplete token in tmpbuf */
		n = preparse.comp_tok_len + 1;
		if (n > sizeof(tmpbuf))
			n = sizeof(tmpbuf);
		snprintf(tmpbuf, n, "%s", preparse.comp_tok_buf);

		/* non completable */
		if (token->ops->complete_start(token, tmpbuf,
					       &preparse.opaque) < 0) {
			if (token->ops->complete_end != NULL)
				token->ops->complete_end(token,
							 &preparse.opaque);
			continue;
		}

		/* all possible completion for this token */
		while (1) {

			ret = token->ops->complete_iterate(token,
							   &preparse.opaque,
							   tmpbuf,
							   sizeof(tmpbuf)-1);

			if (ret < 0)
				break;

			debug_printf("Completion %s\n", tmpbuf);

			/* we kept at least the room for one char */
			if (ret == 0)
				strcat(tmpbuf, " ");

			debug_printf("   choice <%s>\n", tmpbuf);

			/* does the completion match the beginning of
			 * the word ? */
			if (strncmp(preparse.comp_tok_buf, tmpbuf,
				    preparse.comp_tok_len))
				continue;

			/* first one, save the buffer */
			if (nb_completion == 0) {
				completion_len = snprintf(completion_buf,
							sizeof(completion_buf),
							"%s", tmpbuf);
			}
			else {
				n = nb_common_chars(completion_buf, tmpbuf);
				if (n < completion_len)
					completion_len = n;
			}
			nb_completion ++;

			/* we cannot add any char, just display help */
			if (completion_len == preparse.comp_tok_len)
				break;
		}
		if (token->ops->complete_end != NULL)
			token->ops->complete_end(token, &preparse.opaque);

		if (completion_len == preparse.comp_tok_len)
			break;
	}

	debug_printf("nb_completion=%d, completion_len=%d\n",
		     nb_completion, completion_len);

	/* one choice, append chars and return */
	if (nb_completion == 1) {
		snprintf(dst, dstsize, "%s",
			 completion_buf + preparse.comp_tok_len);
		return CMDLINE_COMPLETE_APPEND;
	}

	/* many choices, but starting with same chars: append chars
	 * and return */
	if (nb_completion != 0 && completion_len > preparse.comp_tok_len) {
		if (completion_len >= dstsize)
			completion_len = dstsize - 1;
		strncpy(dst, completion_buf + preparse.comp_tok_len,
			completion_len - preparse.comp_tok_len);
		dst[completion_len - preparse.comp_tok_len] = '\0';
		return CMDLINE_COMPLETE_APPEND;
	}

	/* no match, nothing to do */
	if (nb_match == 0 || nb_completion == 0)
		return CMDLINE_COMPLETE_NONE;

	return CMDLINE_COMPLETE_MANY;
}

