/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <inttypes.h>
#include <ctype.h>
#include <string.h>

#include "cmdline_parse.h"
#include "cmdline_parse_string.h"

#define MULTISTRING_HELP "Mul-choice STRING"
#define ANYSTRING_HELP   "Any STRING"
#define FIXEDSTRING_HELP "Fixed STRING"

static unsigned int
get_token_len(const char *s)
{
	char c;
	unsigned int i=0;

	c = s[i];
	while (c!='#' && c!='\0') {
		i++;
		c = s[i];
	}
	return i;
}

static const char *
get_next_token(const char *s)
{
	unsigned int i;
	i = get_token_len(s);
	if (s[i] == '#')
		return s+i+1;
	return NULL;
}

static int
parse_fixed_string(struct cmdline_token_string_data *sd,
		   const char *buf, unsigned token_len)
{
	unsigned int conf_token_len;
	const char *str;

	str = sd->str;
	for (str = sd->str; str != NULL ; str = get_next_token(str)) {

		conf_token_len = get_token_len(str);

		/* if token from config is too big... */
		if (conf_token_len >= STR_TOKEN_SIZE - 1)
			continue;

		/* compare conf token and user token */
		if (token_len == conf_token_len &&
		    strncmp(buf, str, token_len) == 0)
			return 0;
	}

	return -1;
}

static int
cmdline_parse_string(cmdline_parse_token_hdr_t *tk, const char *buf, void *res,
		     unsigned ressize)
{
	struct cmdline_token_string *tk2 = (struct cmdline_token_string *)tk;
	struct cmdline_token_string_data *sd = &tk2->string_data;;
	unsigned int token_len;

	if (res && ressize < STR_TOKEN_SIZE)
		return -1;

	token_len = strlen(buf);

	if (token_len >= (STR_TOKEN_SIZE - 1) || token_len == 0)
		return -1;

	/* fixed string */
	if (sd->str) {
		if (parse_fixed_string(sd, buf, token_len) < 0)
			return -1;
	}

	/* we already checked that token_len is < STR_TOKEN_SIZE-1 */
	if (res)
		strcpy(res, buf);

	return 0;
}

static int
cmdline_complete_string_start(cmdline_parse_token_hdr_t *tk,
			      __attribute__((unused)) const char *tokstr,
			      void **opaque)
{
	struct cmdline_token_string *tk2 = (struct cmdline_token_string *)tk;
	struct cmdline_token_string_data *sd = &tk2->string_data;;
	const char *str;

	str = sd->str;
	*opaque = (void *)str;
	if (str == NULL)
		return -1; /* no completion */
	return 0;
}

static int
cmdline_complete_string_iterate(cmdline_parse_token_hdr_t *tk, void **opaque,
				char *dstbuf, unsigned int size)
{
	const char *s;
	unsigned int len;

	s = *opaque;
	if (s == NULL)
		return -1;
	*opaque = (void *)get_next_token(s);

	len = get_token_len(s);
	if (len > size - 1)
		return -1;

	memcpy(dstbuf, s, len);
	dstbuf[len] = '\0';
	return 0;
}

static int
cmdline_help_string(cmdline_parse_token_hdr_t *tk, char *dstbuf,
		    unsigned int size)
{
	struct cmdline_token_string *tk2 = (struct cmdline_token_string *)tk;
	struct cmdline_token_string_data *sd = &tk2->string_data;;
	const char *s;

	s = sd->str;

	if (s != NULL) {
		if (get_next_token(s))
			strncpy(dstbuf, MULTISTRING_HELP, size);
		else
			strncpy(dstbuf, FIXEDSTRING_HELP, size);
	}
	else {
		strncpy(dstbuf, ANYSTRING_HELP, size);
	}

	dstbuf[size-1] = '\0';

	return 0;
}

struct cmdline_token_ops cmdline_token_string_ops = {
	.parse = cmdline_parse_string,
	.complete_start = cmdline_complete_string_start,
	.complete_iterate = cmdline_complete_string_iterate,
	.complete_end = NULL,
	.help = cmdline_help_string,
};
