/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>

#include <cirbuf.h>

#include "cmdline_rdline.h"
#include "cmdline_parse.h"

#ifndef NO_RDLINE_HISTORY
static void rdline_remove_old_history_item(struct rdline *rdl);
static void rdline_remove_first_history_item(struct rdline *rdl);
static unsigned int rdline_get_history_size(struct rdline *rdl);
#endif /* !NO_RDLINE_HISTORY */

#ifndef NO_PAGER
static int rdline_pager_next_page(struct rdline *rdl);
static void rdline_asyncpager_reset(struct rdline *rdl);
#endif /* !NO_PAGER */


/* isblank() needs _XOPEN_SOURCE >= 600 || _ISOC99_SOURCE, so use our
 * own. */
static int
isblank2(char c)
{
	if (c == ' ' ||
	    c == '\t' )
		return 1;
	return 0;
}

void
rdline_init(struct rdline *rdl,
	    int fd_in, int fd_out,
	    rdline_validate_t *validate,
	    rdline_complete_t *complete,
	    rdline_help_t *help)
{
	memset(rdl, 0, sizeof(*rdl));
	rdl->fd_in = fd_in;
	rdl->fd_out = fd_out;
	rdl->validate = validate;
	rdl->complete = complete;
	rdl->help = help;
	rdl->status = RDLINE_STOPPED;
#ifndef NO_RDLINE_HISTORY
	cirbuf_init(&rdl->history, rdl->history_buf, 0, RDLINE_HISTORY_BUF_SIZE);
#endif /* !NO_RDLINE_HISTORY */
}

void
rdline_newline(struct rdline *rdl, const char *prompt)
{
	vt100_init(&rdl->vt100);
	cirbuf_init(&rdl->left, rdl->left_buf, 0, RDLINE_BUF_SIZE);
	cirbuf_init(&rdl->right, rdl->right_buf, 0, RDLINE_BUF_SIZE);

	/* if pointer is the same or NULL, don't copy it */
	if (prompt != NULL && prompt != rdl->prompt)
		snprintf(rdl->prompt, sizeof(rdl->prompt), "%s", prompt);

	rdline_printf(rdl, "%s", rdl->prompt);
	rdl->status = RDLINE_RUNNING;

#ifndef NO_RDLINE_HISTORY
	rdl->history_cur_line = -1;
#endif /* !NO_RDLINE_HISTORY */
}

void
rdline_stop(struct rdline *rdl)
{
	rdl->status = RDLINE_STOPPED;
}

void
rdline_quit(struct rdline *rdl)
{
	rdl->status = RDLINE_EXITED;
}

void
rdline_restart(struct rdline *rdl)
{
	rdl->status = RDLINE_RUNNING;
}

const char *
rdline_get_buffer(struct rdline *rdl)
{
	unsigned int len_l, len_r;
	cirbuf_align_left(&rdl->left);
	cirbuf_align_left(&rdl->right);

	len_l = CIRBUF_GET_LEN(&rdl->left);
	len_r = CIRBUF_GET_LEN(&rdl->right);
	memcpy(rdl->left_buf+len_l, rdl->right_buf, len_r);

	rdl->left_buf[len_l + len_r] = '\0';
	return rdl->left_buf;
}

static void
display_right_buffer(struct rdline *rdl, int force)
{
	unsigned int i;
	char tmp;

	if (!force && CIRBUF_IS_EMPTY(&rdl->right))
		return;

	rdline_printf(rdl, vt100_clear_right);
	CIRBUF_FOREACH(&rdl->right, i, tmp) {
		rdline_printf(rdl, "%c", tmp);
	}
	if (!CIRBUF_IS_EMPTY(&rdl->right))
		rdline_printf(rdl, vt100_multi_left,
			      CIRBUF_GET_LEN(&rdl->right));
}

void
rdline_redisplay(struct rdline *rdl)
{
	unsigned int i;
	char tmp;

	rdline_printf(rdl, vt100_home);
	rdline_printf(rdl, "%s", rdl->prompt);
	CIRBUF_FOREACH(&rdl->left, i, tmp) {
		rdline_printf(rdl, "%c", tmp);
	}
	display_right_buffer(rdl, 1);
}

static int
rdline_parse_char(struct rdline *rdl, char c)
{
	unsigned int i;
	int cmd;
	char tmp;
#ifndef NO_RDLINE_HISTORY
	char *buf;
#endif

	cmd = vt100_parser(&rdl->vt100, c);
	if (cmd == VT100_NOT_COMPLETE)
		return RDLINE_RES_SUCCESS;

#ifndef NO_PAGER
	/* display asynchrounous printf if any */
	if (rdl->pager_buf != NULL) {

		/* user ask to exit pager, or last page is displayed*/
		if ((cmd == VT100_STD_CHAR && c == 'q') ||
		    rdline_pager_next_page(rdl) == 0) {
			int ret;

			ret = rdl->pager_ret;
			rdline_asyncpager_reset(rdl);
			if (rdl->pager_cb != NULL) {
				rdl->pager_cb(rdl, rdl->pager_arg);
				rdl->pager_cb = NULL;
			}
			/* maybe the pager was reloaded in the
			 * callback */
			if (rdl->pager_buf != NULL)
				return RDLINE_RES_SUCCESS;

			/* else, redisplay prompt and return the saved status */
			rdline_redisplay(rdl);
			return ret;
		}

		/* Some pages remain, lines were displayed in
		 * rdline_pager_next_page() */
		return RDLINE_RES_SUCCESS;
	}
#endif

	/* process control chars */
	if (cmd != VT100_STD_CHAR) {
		switch (cmd) {
		case CMDLINE_KEY_CTRL_B:
		case CMDLINE_KEY_LEFT_ARR:
			if (CIRBUF_IS_EMPTY(&rdl->left))
				break;
			tmp = cirbuf_get_tail(&rdl->left);
			cirbuf_del_tail(&rdl->left);
			cirbuf_add_head(&rdl->right, tmp);
			rdline_printf(rdl, vt100_left_arr);
			break;

		case CMDLINE_KEY_CTRL_F:
		case CMDLINE_KEY_RIGHT_ARR:
			if (CIRBUF_IS_EMPTY(&rdl->right))
				break;
			tmp = cirbuf_get_head(&rdl->right);
			cirbuf_del_head(&rdl->right);
			cirbuf_add_tail(&rdl->left, tmp);
			rdline_printf(rdl, vt100_right_arr);
			break;

		case CMDLINE_KEY_WLEFT:
			while (! CIRBUF_IS_EMPTY(&rdl->left) &&
			       (tmp = cirbuf_get_tail(&rdl->left)) &&
			       isblank2(tmp)) {
				rdline_printf(rdl, vt100_left_arr);
				cirbuf_del_tail(&rdl->left);
				cirbuf_add_head(&rdl->right, tmp);
			}
			while (! CIRBUF_IS_EMPTY(&rdl->left) &&
			       (tmp = cirbuf_get_tail(&rdl->left)) &&
			       !isblank2(tmp)) {
				rdline_printf(rdl, vt100_left_arr);
				cirbuf_del_tail(&rdl->left);
				cirbuf_add_head(&rdl->right, tmp);
			}
			break;

		case CMDLINE_KEY_WRIGHT:
			while (! CIRBUF_IS_EMPTY(&rdl->right) &&
			       (tmp = cirbuf_get_head(&rdl->right)) &&
			       isblank2(tmp)) {
				rdline_printf(rdl, vt100_right_arr);
				cirbuf_del_head(&rdl->right);
				cirbuf_add_tail(&rdl->left, tmp);
			}
			while (! CIRBUF_IS_EMPTY(&rdl->right) &&
			       (tmp = cirbuf_get_head(&rdl->right)) &&
			       !isblank2(tmp)) {
				rdline_printf(rdl, vt100_right_arr);
				cirbuf_del_head(&rdl->right);
				cirbuf_add_tail(&rdl->left, tmp);
			}
			break;

		case CMDLINE_KEY_BKSPACE:
			if(!cirbuf_del_tail_safe(&rdl->left)) {
				rdline_printf(rdl, vt100_bs);
				display_right_buffer(rdl, 1);
			}
			break;

		case CMDLINE_KEY_META_BKSPACE:
		case CMDLINE_KEY_CTRL_W:
			while (! CIRBUF_IS_EMPTY(&rdl->left) &&
			       isblank2(cirbuf_get_tail(&rdl->left))) {
				rdline_printf(rdl, vt100_bs);
				cirbuf_del_tail(&rdl->left);
			}
			while (! CIRBUF_IS_EMPTY(&rdl->left) &&
			       !isblank2(cirbuf_get_tail(&rdl->left))) {
				rdline_printf(rdl, vt100_bs);
				cirbuf_del_tail(&rdl->left);
			}
			display_right_buffer(rdl, 1);
			break;

		case CMDLINE_KEY_META_D:
			while (! CIRBUF_IS_EMPTY(&rdl->right) &&
			       isblank2(cirbuf_get_head(&rdl->right)))
				cirbuf_del_head(&rdl->right);
			while (! CIRBUF_IS_EMPTY(&rdl->right) &&
			       !isblank2(cirbuf_get_head(&rdl->right)))
				cirbuf_del_head(&rdl->right);
			display_right_buffer(rdl, 1);
			break;

		case CMDLINE_KEY_SUPPR:
		case CMDLINE_KEY_CTRL_D:
			if (cmd == CMDLINE_KEY_CTRL_D &&
			    CIRBUF_IS_EMPTY(&rdl->left) &&
			    CIRBUF_IS_EMPTY(&rdl->right)) {
				return RDLINE_RES_EOF;
			}
			if (!cirbuf_del_head_safe(&rdl->right)) {
				display_right_buffer(rdl, 1);
			}
			break;

		case CMDLINE_KEY_CTRL_A:
			if (CIRBUF_IS_EMPTY(&rdl->left))
				break;
			rdline_printf(rdl, vt100_multi_left,
				      CIRBUF_GET_LEN(&rdl->left));
			while (! CIRBUF_IS_EMPTY(&rdl->left)) {
				tmp = cirbuf_get_tail(&rdl->left);
				cirbuf_del_tail(&rdl->left);
				cirbuf_add_head(&rdl->right, tmp);
			}
			break;

		case CMDLINE_KEY_CTRL_E:
			if (CIRBUF_IS_EMPTY(&rdl->right))
				break;
			rdline_printf(rdl, vt100_multi_right,
				      CIRBUF_GET_LEN(&rdl->right));
			while (! CIRBUF_IS_EMPTY(&rdl->right)) {
				tmp = cirbuf_get_head(&rdl->right);
				cirbuf_del_head(&rdl->right);
				cirbuf_add_tail(&rdl->left, tmp);
			}
			break;

#ifndef NO_RDLINE_KILL_BUF
		case CMDLINE_KEY_CTRL_K:
			cirbuf_get_buf_head(&rdl->right, rdl->kill_buf, RDLINE_BUF_SIZE);
			rdl->kill_size = CIRBUF_GET_LEN(&rdl->right);
			cirbuf_del_buf_head(&rdl->right, rdl->kill_size);
			rdline_printf(rdl, vt100_clear_right);
			break;

		case CMDLINE_KEY_CTRL_Y:
			i=0;
			while(CIRBUF_GET_LEN(&rdl->right) + CIRBUF_GET_LEN(&rdl->left) <
			      RDLINE_BUF_SIZE &&
			      i < rdl->kill_size) {
				cirbuf_add_tail(&rdl->left, rdl->kill_buf[i]);
				rdline_printf(rdl, "%c", rdl->kill_buf[i]);
				i++;
			}
			display_right_buffer(rdl, 0);
			break;
#endif /* !NO_RDLINE_KILL_BUF */

		case CMDLINE_KEY_CTRL_C:
			rdline_printf(rdl, "\r\n");
			rdline_newline(rdl, rdl->prompt);
			break;

		case CMDLINE_KEY_CTRL_L:
			rdline_redisplay(rdl);
			break;

		case CMDLINE_KEY_HELP: {
			if (rdl->help == NULL)
				break;

			cirbuf_align_left(&rdl->left);
			rdl->left_buf[CIRBUF_GET_LEN(&rdl->left)] = '\0';
			rdline_printf(rdl, "\r\n");
#ifndef NO_PAGER
			rdl->help(rdl, rdl->left_buf, rdline_asyncpager_write, rdl);
			if (rdl->pager_buf != NULL)
				return RDLINE_RES_SUCCESS;
			else
				rdline_asyncpager_reset(rdl);
#else
			rdl->help(rdl, rdl->left_buf, rdline_write, rdl);
#endif
			rdline_redisplay(rdl);
			break;
		}

		case CMDLINE_KEY_TAB: {
			char tmp_buf[CMDLINE_MAX_TOKEN_SIZE];
			int ret;
			unsigned int tmp_size;

			if (rdl->complete == NULL)
				break;

			cirbuf_align_left(&rdl->left);
			rdl->left_buf[CIRBUF_GET_LEN(&rdl->left)] = '\0';

			/* see in parse.h for help on complete() */
			ret = rdl->complete(rdl, rdl->left_buf,
					    tmp_buf, sizeof(tmp_buf));
			/* no completion or error */
			if (ret == CMDLINE_COMPLETE_NONE ||
			    ret == CMDLINE_COMPLETE_MANY) {
				cirbuf_align_left(&rdl->left);
				rdl->left_buf[CIRBUF_GET_LEN(&rdl->left)] = '\0';
				rdline_printf(rdl, "\r\n");
#ifndef NO_PAGER
				rdl->help(rdl, rdl->left_buf, rdline_asyncpager_write,
					  rdl);
				if (rdl->pager_buf != NULL)
					return RDLINE_RES_SUCCESS;
				else
					rdline_asyncpager_reset(rdl);
#else
				rdl->help(rdl, rdl->left_buf, rdline_write, rdl);
#endif
				rdline_redisplay(rdl);
				break;
			}

			tmp_size = strlen(tmp_buf);
			/* add chars */
			i = 0;
			while(CIRBUF_GET_LEN(&rdl->right) + CIRBUF_GET_LEN(&rdl->left) <
			      RDLINE_BUF_SIZE &&
			      i < tmp_size) {
				cirbuf_add_tail(&rdl->left, tmp_buf[i]);
				rdline_printf(rdl, "%c", tmp_buf[i]);
				i++;
			}
			display_right_buffer(rdl, 1);
			break;
		}

		case CMDLINE_KEY_RETURN:
		case CMDLINE_KEY_RETURN2: {
			char tmp;
			while (!CIRBUF_IS_EMPTY(&rdl->right) &&
			       (tmp = cirbuf_get_head(&rdl->right))) {
				cirbuf_del_head(&rdl->right);
				cirbuf_add_tail(&rdl->left, tmp);
			}
			cirbuf_align_left(&rdl->left);
			rdl->left_buf[CIRBUF_GET_LEN(&rdl->left)] = '\0';
			rdline_printf(rdl, "\r\n");
#ifndef NO_RDLINE_HISTORY
			if (rdl->history_cur_line != -1)
				rdline_remove_first_history_item(rdl);
#endif

			if (rdl->validate)
				rdl->validate(rdl, rdl->left_buf,
					      CIRBUF_GET_LEN(&rdl->left));
#ifndef NO_PAGER
			/* user may have stopped rdline */
			if (rdl->status == RDLINE_EXITED) {
				rdline_asyncpager_reset(rdl);
				return RDLINE_RES_EXITED;
			}
			/* there is something in pager buffer, save
			 * return value that will be return once
			 * paging is finished */
			if (rdl->pager_buf != NULL) {
				rdl->pager_ret = RDLINE_RES_VALIDATED;
				return RDLINE_RES_SUCCESS;
			}

			rdline_asyncpager_reset(rdl);
#else
			if (rdl->status == RDLINE_EXITED)
				return RDLINE_RES_EXITED;
#endif
			return RDLINE_RES_VALIDATED;
		}
#ifndef NO_RDLINE_HISTORY
		case CMDLINE_KEY_UP_ARR:
		case CMDLINE_KEY_CTRL_P:
			if (rdl->history_cur_line == 0) {
				rdline_remove_first_history_item(rdl);
			}
			if (rdl->history_cur_line <= 0) {
				rdline_add_history(rdl, rdline_get_buffer(rdl));
				rdl->history_cur_line = 0;
			}

			buf = rdline_get_history_item(rdl, rdl->history_cur_line + 1);
			if (!buf)
				break;

			rdl->history_cur_line ++;
			vt100_init(&rdl->vt100);
			cirbuf_init(&rdl->left, rdl->left_buf, 0, RDLINE_BUF_SIZE);
			cirbuf_init(&rdl->right, rdl->right_buf, 0, RDLINE_BUF_SIZE);
			cirbuf_add_buf_tail(&rdl->left, buf, strlen(buf));
			rdline_redisplay(rdl);
			break;

		case CMDLINE_KEY_DOWN_ARR:
		case CMDLINE_KEY_CTRL_N:
			if (rdl->history_cur_line - 1 < 0)
				break;

			rdl->history_cur_line --;
			buf = rdline_get_history_item(rdl, rdl->history_cur_line);
			if (!buf)
				break;
			vt100_init(&rdl->vt100);
			cirbuf_init(&rdl->left, rdl->left_buf, 0, RDLINE_BUF_SIZE);
			cirbuf_init(&rdl->right, rdl->right_buf, 0, RDLINE_BUF_SIZE);
			cirbuf_add_buf_tail(&rdl->left, buf, strlen(buf));
			rdline_redisplay(rdl);

			break;
#endif /* !NO_RDLINE_HISTORY */

		default:
			break;
		}

		return RDLINE_RES_SUCCESS;
	}

	if (!isprint((int)c))
		return RDLINE_RES_SUCCESS;

	/* standard chars */
	if (CIRBUF_GET_LEN(&rdl->left) + CIRBUF_GET_LEN(&rdl->right) >= RDLINE_BUF_SIZE)
		return RDLINE_RES_SUCCESS;

	if (cirbuf_add_tail_safe(&rdl->left, c))
		return RDLINE_RES_SUCCESS;

	rdline_printf(rdl, "%c", c);
	display_right_buffer(rdl, 0);

	return RDLINE_RES_SUCCESS;
}

int
rdline_char_in(struct rdline *rdl, char c)
{
	int ret, same = 0;
	const char *history, *buffer;

	if (rdl->status == RDLINE_EXITED)
		return RDLINE_RES_EXITED;
	if (rdl->status != RDLINE_RUNNING)
		return RDLINE_RES_NOT_RUNNING;

	ret = rdline_parse_char(rdl, c);

	/* add line to history */
	if (ret == RDLINE_RES_VALIDATED) {
		buffer = rdline_get_buffer(rdl);
		history = rdline_get_history_item(rdl, 0);
		if (history)
			same = !strcmp(buffer, history);

		if (strlen(buffer) >= 1 && same == 0)
			rdline_add_history(rdl, buffer);
	}

	return ret;
}

int
rdline(struct rdline *rdl, const char *prompt)
{
	char c;
	int ret = RDLINE_RES_NOT_RUNNING;

	rdline_newline(rdl, prompt);
	while (1) {
		if (read(rdl->fd_in, &c, 1) < 0)
			break;
		ret = rdline_char_in(rdl, c);
		if (ret != RDLINE_RES_SUCCESS)
			break;
	}

	return ret;
}

/* HISTORY */

#ifndef NO_RDLINE_HISTORY
static void
rdline_remove_old_history_item(struct rdline * rdl)
{
	char tmp;

	while (! CIRBUF_IS_EMPTY(&rdl->history) ) {
		tmp = cirbuf_get_head(&rdl->history);
		cirbuf_del_head(&rdl->history);
		if (!tmp)
			break;
	}
}

static void
rdline_remove_first_history_item(struct rdline * rdl)
{
	char tmp;

	if ( CIRBUF_IS_EMPTY(&rdl->history) ) {
		return;
	}
	else {
		cirbuf_del_tail(&rdl->history);
	}

	while (! CIRBUF_IS_EMPTY(&rdl->history) ) {
		tmp = cirbuf_get_tail(&rdl->history);
		if (!tmp)
			break;
		cirbuf_del_tail(&rdl->history);
	}
}

static unsigned int
rdline_get_history_size(struct rdline * rdl)
{
	unsigned int i, tmp, ret=0;

	CIRBUF_FOREACH(&rdl->history, i, tmp) {
		if (tmp == 0)
			ret ++;
	}

	return ret;
}

char *
rdline_get_history_item(struct rdline * rdl, unsigned int idx)
{
	unsigned int len, i, tmp;

	len = rdline_get_history_size(rdl);
	if ( idx >= len ) {
		return NULL;
	}

	cirbuf_align_left(&rdl->history);

	CIRBUF_FOREACH(&rdl->history, i, tmp) {
		if ( idx == len - 1) {
			return rdl->history_buf + i;
		}
		if (tmp == 0)
			len --;
	}

	return NULL;
}

int
rdline_add_history(struct rdline * rdl, const char * buf)
{
	unsigned int len;

	len = strlen(buf);
	if (len >= RDLINE_HISTORY_BUF_SIZE)
		return -1;

	while (len >= CIRBUF_GET_FREELEN(&rdl->history)) {
		rdline_remove_old_history_item(rdl);
	}

	cirbuf_add_buf_tail(&rdl->history, buf, len);
	cirbuf_add_tail(&rdl->history, 0);

	return 0;
}

void
rdline_clear_history(struct rdline * rdl)
{
	cirbuf_init(&rdl->history, rdl->history_buf, 0, RDLINE_HISTORY_BUF_SIZE);
}

#else /* !NO_RDLINE_HISTORY */

int rdline_add_history(struct rdline * rdl, const char * buf)
{
	return -1;
}

void rdline_clear_history(struct rdline * rdl)
{
	return;
}

char * rdline_get_history_item(struct rdline * rdl, unsigned int i)
{
	return NULL;
}


#endif /* !NO_RDLINE_HISTORY */


ssize_t
rdline_write(const struct rdline *rdl, void *buf, size_t count)
{
	return write(rdl->fd_out, buf, count);
}

int
rdline_vprintf(const struct rdline *rdl, const char *fmt, va_list ap)
{
	int ret;
#ifndef _GNU_SOURCE
	char *buf;
#endif

	if (rdl->fd_out < 0)
		return -1;

#ifdef _GNU_SOURCE
	ret = vdprintf(rdl->fd_out, fmt, ap);
#else
	buf = malloc(BUFSIZ);
	if (buf == NULL)
		return -1;

	ret = vsnprintf(buf, BUFSIZ, fmt, ap);

	if (ret > 0)
		write(rdl->fd_out, buf, (ret >= BUFSIZ) ? BUFSIZ : ret);
	free(buf);
#endif

	return ret;
}

int
rdline_printf(const struct rdline *rdl, const char *fmt, ...)
{
	va_list ap;
	int ret;

	va_start(ap, fmt);
	ret = rdline_vprintf(rdl, fmt, ap);
	va_end(ap);

	return ret;
}

#ifndef NO_PAGER
/* reset pager state */
void
rdline_asyncpager_reset(struct rdline *rdl)
{
	if (rdl->pager_buf) {
		free(rdl->pager_buf);
		rdl->pager_buf = NULL;
	}
	rdl->pager_lines = 0;
	rdl->pager_len = 0;
	rdl->pager_off = 0;
	rdl->pager_cb = NULL;
	rdl->pager_arg = NULL;
	rdl->pager_ret = RDLINE_RES_SUCCESS;
}

/* Return the offset of the i-th occurence of char c in string s. If
 * there is less than i occurences, return -1 and fill i with the
 * count. */
static int
strnchr(const char *s, char c, int *i)
{
	int n = 0;
	const char *orig = s;

	while (*s) {
		if (*s == c)
			n++;
		if (*i == n)
			return s - orig;
		s++;
	}
	*i = n;
	return -1;
}

/* display a page of data from pager, return 0 if all is displayed */
static int
rdline_pager_next_page(struct rdline *rdl)
{
	int lines = RDLINE_MAX_LINES;
	int displen;
	char *s;

	s = rdl->pager_buf;
	if (s == NULL)
		return 0;

	rdline_printf(rdl, vt100_home);
	rdline_printf(rdl, vt100_clear_right);

	s += rdl->pager_off;

	/* we know that s is 0-terminated */
	displen = strnchr(s, '\n', &lines);
	rdl->pager_lines = lines;

	/* we can display all the data */
	if (displen == -1) {
		write(rdl->fd_out, s, rdl->pager_len);
		free(rdl->pager_buf);
		rdl->pager_buf = NULL;
		return 0;
	}

	displen = displen + 1; /* include \n */
	write(rdl->fd_out, s, displen);
	rdl->pager_off += displen;
	rdl->pager_len -= displen;

	rdline_printf(rdl, "--- press a key to continue ---");
	return -1;
}

/* push data in pager */
ssize_t
rdline_asyncpager_write(struct rdline *rdl, void *buf, size_t len)
{
	char *s = buf;

	/* display as many lines as we can */
	if (rdl->pager_lines < RDLINE_MAX_LINES) {
		int lines = RDLINE_MAX_LINES - rdl->pager_lines;
		int displen;

		/* we know that s is 0-terminated */
		displen = strnchr(s, '\n', &lines);
		rdl->pager_lines += lines;

		/* we can display all the data */
		if (displen == -1) {
			write(rdl->fd_out, s, len);
			return 0;
		}
		displen = displen + 1; /* include \n */
		write(rdl->fd_out, s, displen);
		s += displen;
		len -= displen;
	}

	if (rdl->pager_buf == NULL) {
		rdline_printf(rdl, "--- press a key to continue ---");
	}
	rdl->pager_buf = realloc(rdl->pager_buf, rdl->pager_len + len);
	if (rdl->pager_buf == NULL) {
		rdline_asyncpager_reset(rdl);
		return -1;
	}

	memcpy(rdl->pager_buf + rdl->pager_len, s, len);
	rdl->pager_len += len;
	return 0;
}

/* Print data asynchronously (using pager if needed) */
int
rdline_asyncpager_printf(struct rdline *rdl, const char *fmt, ...)
{
	int n;
	char *buf;
	va_list ap;

	if (rdl->fd_out < 0)
		return -1;

	buf = malloc(BUFSIZ);
	if (buf == NULL)
		return -1;

	va_start(ap, fmt);
	n = vsnprintf(buf, BUFSIZ, fmt, ap);
	va_end(ap);

	if (n >= BUFSIZ)
		n = BUFSIZ-1;
	if (n > 0)
		rdline_asyncpager_write(rdl, buf, n);
	free(buf);
	return n;
}

int rdline_asyncpager_set_cb(struct rdline *rdl, rdline_asyncpager_cb_t *cb,
			     void *arg)
{
	if (rdl->pager_buf == NULL)
		return -1;

	rdl->pager_cb = cb;
	rdl->pager_arg = arg;
	return 0;
}
#endif /* !NO_PAGER */
