/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CMDLINE_PARSE_H_
#define CMDLINE_PARSE_H_

#include <sys/types.h>

#ifndef offsetof
#define offsetof(type, field)  ((size_t) &( ((type *)0)->field) )
#endif

#define CMDLINE_MAX_TOKEN_SIZE 128 /* including '\0' */
#define CMDLINE_MAX_DSTBUF_SIZE 1024

/**
 * Stores a pointer to the ops struct, and the offset: the place to
 * write the parsed result in the destination structure.
 */
struct cmdline_token_hdr {
	struct cmdline_token_ops *ops;
	unsigned int offset;
};
typedef struct cmdline_token_hdr cmdline_parse_token_hdr_t;

/**
 * A token is defined by this structure.
 *
 * parse() takes the token as first argument, then the source buffer
 * containing the token we want to parse. The 3rd arg is a pointer
 * where we store the parsed data (as binary), and the 4th arg is the
 * size of this area. It returns 0 on success and a negative value on
 * error.
 *
 * complete_start() prepares a completion operation. The first
 * argument is the token to complete. The second argument is the token
 * to complete, and the third arg is an opaque pointer that will be
 * given to complete_iterate() function. It can be used to store
 * private data for this completion. For each complete_start() call,
 * the user must call complete_end() at the end of iterations (if
 * defined). Return a negative value if completion is not possible, or
 * 0 on success.
 *
 * complete_iterate() copy in dstbuf (the size is specified in the
 * parameter) the next possible completion for this token. Return 0 on
 * success (final completion), 1 if it's an intermediate completion,
 * or a negative value on error (or when there is no more
 * completion). Refer to cmdline_complete_string_iterate() for an
 * example.
 *
 * complete_end() is called when the iteration on this token is finished,
 * this function should free all things allocated during complete_start().
 *
 * help() fills the dstbuf with the help for the token. It returns
 * -1 on error and 0 on success.
 */
struct cmdline_token_ops {
	/** parse(token ptr, buf, res pts) */
	int (*parse)(cmdline_parse_token_hdr_t *, const char *, void *,
		     unsigned int);
	/** prepare a completion on this token */
	int (*complete_start)(cmdline_parse_token_hdr_t *, const char *,
			      void **);
	/** fill dstbuf for this token (token, opaque, dstbuf, size) */
	int (*complete_iterate)(cmdline_parse_token_hdr_t *, void **, char *,
				unsigned int);
	/* end of completion, used to free the opaque structure */
	void (*complete_end)(cmdline_parse_token_hdr_t *, void **);
	/** get help for this token (token, dstbuf, size) */
	int (*help)(cmdline_parse_token_hdr_t *, char *, unsigned int);
};

struct cmdline;
/**
 * Store a instruction, which is a pointer to a callback function and
 * its parameter that is called when the instruction is parsed, a help
 * string, and a list of token composing this instruction.
 */
struct cmdline_inst {
	/* f(parsed_struct, data) */
	void (*f)(void *, struct cmdline *, void *);
	void *data;
	char *help_str;
	cmdline_parse_token_hdr_t *tokens[];
};
typedef struct cmdline_inst cmdline_parse_inst_t;

/**
 * A context is identified by its name, and contains a list of
 * instruction
 *
 */
struct cmdline_parse_ctx {
	const char *name;
	const cmdline_parse_inst_t *insts[];
};
typedef struct cmdline_parse_ctx cmdline_parse_ctx_t;

/* return status for parsing */
#define CMDLINE_PARSE_SUCCESS        0
#define CMDLINE_PARSE_EMPTY         -1
#define CMDLINE_PARSE_NOMATCH       -2
#define CMDLINE_PARSE_AMBIGUOUS     -3
#define CMDLINE_PARSE_UNTERMINATED_QUOTE -4

/**
 * Try to parse a buffer according to the specified context. The
 * argument linebuf must end with "\n\0".
 *
 * The function returns:
 * - CMDLINE_PARSE_SUCCESS (0) on success
 * - CMDLINE_PARSE_EMPTY if there is nothing to parse
 * - CMDLINE_PARSE_NOMATCH if line does not match any command
 * - CMDLINE_PARSE_AMBIGUOUS if several commands match
 * - CMDLINE_PARSE_UNTERMINATED_QUOTE if a quote is used incorrectly
 */
int cmdline_parse(cmdline_parse_ctx_t *ctx, const char *linebuf, void *opaque);

/* return status for completion */
#define CMDLINE_COMPLETE_APPEND    0
#define CMDLINE_COMPLETE_NONE     -1
#define CMDLINE_COMPLETE_MANY     -2

/**
 * cmdline_complete() tries to complete the buffer given as a parameter.
 *
 * It returns:
 *   - CMDLINE_COMPLETE_APPEND (0) on success, when a completion is
 *     done (one possible choice). In this case, the chars are
 *     appended in dst buffer.
 *   - CMDLINE_COMPLETE_NONE: error, no possible completion
 *   - CMDLINE_COMPLETE_MANY: error, many possble completions, need to call
 *     cmdline_help() function to see all the possibilities.
 */
int cmdline_complete(cmdline_parse_ctx_t *ctx, const char *buf,
		     char *dst, unsigned int size);

/**
 * callback given to rdline_help() to display the content of the
 * help. The first argument is an opaque pointer. The other args
 * are buffer and size.
 */
typedef ssize_t (cmdline_write_t)(void *, void *, size_t);

/**
 * Display a contextual help using the write_buf() function pointer
 * given as parameter (called with its opaque pointer). The contextual
 * help depends on the buffer given.
 */
int cmdline_help(cmdline_parse_ctx_t *ctx, const char *buf,
		 cmdline_write_t *write_buf, void *opaque);

/* return true if(!c || iscomment(c) || isblank(c) ||
 * isendofline(c)) */
int cmdline_isendoftoken(char c);

/* quote a string and escape original quotes */
int cmdline_quote_token(char *dst, unsigned dstlen, const char *src);

/* remove quote and stop when we reach the end of token */
int cmdline_unquote_token(char *dst, unsigned dstlen, const char *src);

#endif /* CMDLINE_PARSE_H_ */
