/*
 * Copyright 2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/queue.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <reent.h>
#include <errno.h>

#include "gloss_device.h"

/* provide a dummy environment */
char *__env[1] = { 0 };
char **environ = __env;

/* provide a definition of errno if not already provided */
int errno;

/* defined by the linker, used by _sbrk */
extern char _ebss;

/* open a device, return its file descriptor */
int
_open_r(struct _reent *reent, const char *file, int flags, int mode)
{
	const struct libc_device *dev;

	dev = libc_device_lookup_from_name(file);
	if (dev == NULL) {
		reent->_errno = ENODEV;
		return -1;
	}

	if (dev->open_r)
		dev->open_r(reent, file, flags, mode);

	return dev->fd;
}

/* close a device */
int
_close_r(struct _reent *reent, int fd)
{
	const struct libc_device *dev;

	dev = libc_device_lookup_from_fd(fd);
	if (dev == NULL) {
		reent->_errno = EBADF;
		return -1;
	}

	/* call the close() handler if defined */
	if (dev->close_r)
		return dev->close_r(reent, fd);

	return 0;
}

/* read a device. */
_ssize_t
_read_r(struct _reent *reent, int fd, void *ptr, size_t len)
{
	const struct libc_device *dev;

	dev = libc_device_lookup_from_fd(fd);
	if (dev == NULL) {
		reent->_errno = EBADF;
		return -1;
	}

	if (dev->read_r)
		return dev->read_r(reent, fd, ptr, len);

	reent->_errno = ENODEV;
	return -1;
}

/* write to a device */
_ssize_t
_write_r(struct _reent *reent, int fd, const void *ptr, size_t len)
{
	const struct libc_device *dev;

	dev = libc_device_lookup_from_fd(fd);
	if (dev == NULL) {
		reent->_errno = EBADF;
		return -1;
	}

	if (dev->write_r)
		return dev->write_r(reent, fd, ptr, len);

	reent->_errno = ENODEV;
	return -1;
}

/* exit: print error code and loop forever */
void
_exit(int rc)
{
	printf("exit %d\n", rc);
	while (1);
}

/* set owner: not implemented */
int
_chown_r(struct _reent *reent,
	 __attribute__((unused)) const char *path,
	 __attribute__((unused)) uid_t owner,
	 __attribute__((unused)) gid_t group)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* execve: not implemented */
int
_execve_r(struct _reent *reent,
	  __attribute__((unused)) const char *name,
	  __attribute__((unused)) char *const *argv,
	  __attribute__((unused)) char *const *env)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* fork: not implemented */
int
_fork_r(struct _reent *reent)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* fstat: not implemented, all devices are character devices */
int
_fstat_r(__attribute__((unused)) struct _reent *reent,
	 __attribute__((unused)) int fildes,
	 struct stat *st)
{
	st->st_mode = S_IFCHR;
	return 0;
}

/* getpid: not implemented */
int
_getpid_r(struct _reent *reent)
{
	reent->_errno = ENOSYS;
	return -1;
}

struct timeval;

/* gettimeofday: not implemented */
int
_gettimeofday_r(struct _reent *reent,
		__attribute__((unused)) struct timeval *ptimeval,
		__attribute__((unused)) void *ptimezone)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* isatty; not implemented */
int
_isatty_r(struct _reent *reent, __attribute__((unused)) int file)
{
	reent->_errno = ENOTTY;
	return 0;
}

/* kill a process: not implemented */
int
_kill_r(struct _reent *reent,
	__attribute__((unused)) int pid,
	__attribute__((unused)) int sig)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* link: not implemented */
int
_link_r(struct _reent *reent,
	__attribute__((unused)) const char *existing,
	__attribute__((unused)) const char *new)
{
	reent->_errno = EMLINK;
	return -1;
}

/* lseek: not implemented */
_off_t
_lseek_r(__attribute__((unused)) struct _reent *reent,
	 __attribute__((unused)) int file,
	 __attribute__((unused)) _off_t ptr,
	 __attribute__((unused)) int dir)
{
	return 0;
}

/* readlink: not implemented */
int
_readlink_r(struct _reent *reent,
	    __attribute__((unused)) const char *path,
	    __attribute__((unused)) char *buf,
	    __attribute__((unused)) size_t bufsize)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* simple sbrk */
void *
_sbrk_r(__attribute__((unused)) struct _reent *reent, ptrdiff_t incr)
{
	static char *heap_end = NULL;
	char *prev_heap_end;

	if (heap_end == NULL)
		heap_end = &_ebss;

	prev_heap_end = heap_end;
	heap_end += incr;
	return (caddr_t)prev_heap_end;
}

/* stat: not implemented, all devices are character devices */
int
_stat_r(__attribute__((unused)) struct _reent *reent,
	__attribute__((unused)) const char *file,
	struct stat *st)
{
	st->st_mode = S_IFCHR;
	return 0;
}

/* symlink: not implemented */
int
_symlink_r(struct _reent *reent,
	   __attribute__((unused)) const char *path1,
	   __attribute__((unused)) const char *path2)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* times: not timplemented */
clock_t
_times_r(struct _reent *reent,
	 __attribute__((unused)) struct tms *buf)
{
	reent->_errno = ENOSYS;
	return -1;
}

/* unlink: not implemented */
int
_unlink_r(struct _reent *reent,
	  __attribute__((unused)) const char *name)
{
	reent->_errno = EMLINK;
	return -1;
}

/* wait: not implemented */
int
_wait_r(struct _reent *reent, __attribute__((unused)) int *status)
{
	reent->_errno = ENOSYS;
	return -1;
}

