/*
 * Copyright 2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef GLOSS_DEVICE_H_
#define GLOSS_DEVICE_H_

#include <sys/queue.h>
#include <string.h>

/**
 * Maximum number of devices stored in cache.
 *
 * Instead of only storing the devices in a list, the first ones are
 * stored in a table indexed by the file descriptor. This avoids a
 * linear browsing on lookup.
 */
#define GLOSS_DEVICE_CACHE_MAX 4

/**
 * Structure describing a device
 *
 * It can be registered using libc_device_register(), and allows to
 * redirect file operations to any channel (uart, spi, ...).
 */
struct libc_device {
	/* filled by libc_device_register() */
	TAILQ_ENTRY(libc_device) next;
	int fd;

	/* filled by user before registration */
	const char *name;
	int (*open_r)(struct _reent *r, const char *path, int flags, int mode);
	int (*close_r)(struct _reent *r, int fd);
	_ssize_t (*read_r)(struct _reent *r, int fd, void *ptr, size_t len);
	_ssize_t (*write_r)(struct _reent *r, int fd, const void *ptr, size_t len);
};

/**
 * The list of devices.
 */
TAILQ_HEAD(libc_device_list, libc_device);
struct libc_device_list libc_device_list;

/**
 * Table of first libc_device pointers, initialized to NULL.
 */
struct libc_device *libc_device_cache[GLOSS_DEVICE_CACHE_MAX];

/**
 * Register a device
 *
 * Register device operations. The structure is appended to the device
 * list, and its name and fd fields are modified. Therefore, a structure
 * cannot be registered twice.
 *
 * The first three devices to be registered are stdin, stdout and stderr.
 *
 * @param dev
 *   The filled device structure
 * @return
 *   0 on success, negative on error
 */
int libc_device_register(struct libc_device *dev);

/**
 * Return the device associated to the given file descriptor.
 *
 * @param fd
 *   The file descriptor.
 * @return
 *   The device matching the file descriptor.
 */
static inline const struct libc_device *
libc_device_lookup_from_fd(int fd)
{
	const struct libc_device *dev;

	if (fd >= 0 && fd < GLOSS_DEVICE_CACHE_MAX)
		return libc_device_cache[fd];

	TAILQ_FOREACH(dev, &libc_device_list, next) {
		if (dev->fd == fd)
			return dev;
	}
	return NULL;
}

/**
 * Return the device associated to the given name.
 *
 * @param name
 *   The name of the file.
 * @return
 *   The device matching the file name.
 */
static inline const struct libc_device *
libc_device_lookup_from_name(const char *name)
{
	const struct libc_device *dev;

	TAILQ_FOREACH(dev, &libc_device_list, next) {
		if (strcmp(name, dev->name) == 0)
			return dev;
	}
	return NULL;
}

#endif /* GLOSS_DEVICE_H_ */
