/*
 * Copyright (c) 2005-2015, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UCG_UART_H_
#define UCG_UART_H_

#include <ucg_cirbuf.h>

typedef void (ucg_uart_cb_t)(char);

struct ucg_uart;

/** this structure stores the configuration of the uart */
struct ucg_uart_config {
	uint8_t enable     : 1, /**< enable or disable the uart */
		parity     : 2, /**< none, odd or even */
		stop_bits  : 1, /**< 1 or 2 bits at the end of the frame */
		reserved   : 4; /**< nothing for now */
	uint8_t nbits;          /**< number of bits in frame, 5,6,7 or 8 */
	uint32_t baudrate;      /**< speed of uart */
};

struct ucg_uart_driver_ops {
	void (*disable_tx_irq)(struct ucg_uart *uart);
	void (*enable_tx_irq)(struct ucg_uart *uart);
	uint8_t (*tx_ready)(struct ucg_uart *uart);
	void (*set_udr)(struct ucg_uart *uart, char c);
	char (*get_udr)(struct ucg_uart *uart);
	int (*setconf)(struct ucg_uart *uart,
		const struct ucg_uart_config *conf);
};

struct ucg_uart {
	struct ucg_uart_config conf;
	ucg_uart_cb_t *rx_cb;
	ucg_uart_cb_t *tx_cb;

	struct ucg_cirbuf *rx_fifo;
	struct ucg_cirbuf *tx_fifo;
	const struct ucg_uart_driver_ops *ops;
	void *driver_data;
};

void ucg_uart_send_next_char(struct ucg_uart *uart);
void ucg_uart_recv_next_char(struct ucg_uart *uart);

/**
 * Initialisation function. This function puts the registers of the
 * microcontroler in a correct state in order to use the uart.  It
 * uses the configuration file <uart_config.h> ; this function is
 * equivalent to call uartX_setconf(NULL) for each uart.
 */
void ucg_uart_init(struct ucg_uart *uart,
	const struct ucg_uart_driver_ops *ops, void *driver_data,
	struct ucg_cirbuf *rx_cbuf, char *rx_buf, unsigned rx_bufsize,
	struct ucg_cirbuf *tx_cbuf, char *tx_buf, unsigned tx_bufsize);

/**
 * Configure the uart with the given configuration. Returns 0 on
 * success.
 */
int ucg_uart_setconf(struct ucg_uart *uart, const struct ucg_uart_config *conf);

/**
 * Get the current configuration of the uart
 */
void ucg_uart_getconf(struct ucg_uart *uart, struct ucg_uart_config *conf);

/**
 * uart_recv returns the next character, taken from the fifo (if
 * any). If there is nothing to read, the function waits until
 * something come on the uart.
 */
int ucg_uart_recv(struct ucg_uart *uart);

/**
 * uart_recv returns the next character, taken from the fifo (if
 * any). If there is nothing to read, the function returns -1.
 */
int ucg_uart_recv_nowait(struct ucg_uart *uart);

/**
 * uart_send_nowait is used to send data to the uart. The data
 * is first stored in the FIFO before beeing sent. If the FIFO is
 * full, data is dropped and the function returns -1, else it returns
 * the character c.
 */
int ucg_uart_send_nowait(struct ucg_uart *uart, char c);

/**
 * uart_send is used to send data to the uart. The data is first
 * stored in the FIFO before beeing sent. If the FIFO is full, the
 * function wait until the uart is ready. The function returns c.
 */
int ucg_uart_send(struct ucg_uart *uart, char c);

/**
 * This function is used to register another function which will be
 * executed at each byte transmission (5, 6 ,7 ,8 bits)
 */
void ucg_uart_register_tx_cb(struct ucg_uart *uart, ucg_uart_cb_t *f);

/**
 * This function is used to register another function which will be
 * executed at each byte reception (5, 6 ,7 ,8 bits)
 */
void ucg_uart_register_rx_cb(struct ucg_uart *uart, ucg_uart_cb_t *f);

#endif /* UCG_UART_H_ */
