/*
 * Copyright (c) 2005-2015, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>

#include <ucg_irq.h>
#include <ucg_cirbuf.h>
#include <ucg_uart.h>

/**
 * transmit next character of fifo if any, and call the event function.
 * This function is executed with intr locked.
 */
void ucg_uart_send_next_char(struct ucg_uart *uart)
{
	char elt = 0;

	if (ucg_cirbuf_is_empty(uart->tx_fifo)) {
		uart->ops->disable_tx_irq(uart);
		return;
	}

	elt = ucg_cirbuf_get_tail(uart->tx_fifo);
	ucg_cirbuf_del_tail(uart->tx_fifo);
	uart->ops->set_udr(uart, elt);
	uart->ops->enable_tx_irq(uart);
}

/**
 * UART RX Interrupt
 */
void ucg_uart_recv_next_char(struct ucg_uart *uart)
{
	char elt = 0;

	elt = uart->ops->get_udr(uart);
	if (!ucg_cirbuf_is_full(uart->rx_fifo)) {
		ucg_cirbuf_add_head(uart->rx_fifo, elt);
	}
	if (uart->rx_cb != NULL)
		uart->rx_cb(elt);
}

/* This function is used to register another function which will be */
/* executed at each byte transmission. */
void ucg_uart_register_tx_cb(struct ucg_uart *uart, void (*f)(char))
{
	ucg_irqflags_t flags;

	flags = ucg_irq_lock_save();
	uart->tx_cb = f;
	ucg_irq_unlock_restore(flags);
}

/* This function is used to register another function which will be */
/* executed at each byte reception */
void ucg_uart_register_rx_cb(struct ucg_uart *uart, void (*f)(char))
{
	ucg_irqflags_t flags;

	flags = ucg_irq_lock_save();
	uart->rx_cb = f;
	ucg_irq_unlock_restore(flags);
}


/* get a char from the receive fifo */
int ucg_uart_recv_nowait(struct ucg_uart *uart)
{
	char elt = 0;
	ucg_irqflags_t flags;

	flags = ucg_irq_lock_save();
	if( !ucg_cirbuf_is_empty(uart->rx_fifo) ) {
		elt = ucg_cirbuf_get_tail(uart->rx_fifo);
		ucg_cirbuf_del_tail(uart->rx_fifo);
		ucg_irq_unlock_restore(flags);
		return (int)elt;
	}
	ucg_irq_unlock_restore(flags);
	return -1;
}

/* get a char from the receive fifo */
int ucg_uart_recv(struct ucg_uart *uart)
{
	int elt = 0;
	while ((elt = ucg_uart_recv_nowait(uart)) == -1)
		;
	return elt;
}

int ucg_uart_send(struct ucg_uart *uart, char c)
{
	/* try to send the char */
	if (ucg_uart_send_nowait(uart, c) != -1)
		return 0;

	/* if it fails, if irq are masked we have to poll */
	if (ucg_irq_locked()) {
		while (!uart->ops->tx_ready(uart))
			;
		/* send data to free room in fifo */
		ucg_uart_send_next_char(uart);
		ucg_cirbuf_add_head(uart->tx_fifo, c);
	}
	else {
		/* if irq are not locked, we can loop to emit */
		while (ucg_uart_send_nowait(uart, c) == -1)
			;
	}
	return 0;
}

/* send a char, or put it in the fifo if uart is not ready. Return -1
 * if fifo is full */
int ucg_uart_send_nowait(struct ucg_uart *uart, char c)
{
	ucg_irqflags_t flags;

	flags = ucg_irq_lock_save();

	if (ucg_cirbuf_is_full(uart->tx_fifo)) {
		ucg_irq_unlock_restore(flags);
		return -1;
	}

	/* uart is ready to send */
	if (ucg_cirbuf_is_empty(uart->tx_fifo) && uart->ops->tx_ready(uart)) {
		uart->ops->set_udr(uart, c);
		uart->ops->enable_tx_irq(uart);
	}
	else { /* not ready, put char in fifo */
		ucg_cirbuf_add_head(uart->tx_fifo, c);
	}

	ucg_irq_unlock_restore(flags);
	return 0;
}

void ucg_uart_getconf(struct ucg_uart *uart, struct ucg_uart_config *conf)
{
	memcpy(conf, &uart->conf, sizeof(*conf));
}

int ucg_uart_setconf(struct ucg_uart *uart, const struct ucg_uart_config *conf)
{
	if (uart->ops->setconf(uart, conf) < 0)
		return -1;
	memcpy(&uart->conf, conf, sizeof(*conf));
	return 0;
}

/* init all uarts */
void ucg_uart_init(struct ucg_uart *uart,
	const struct ucg_uart_driver_ops *ops, void *driver_data,
	struct ucg_cirbuf *rx_cbuf, char *rx_buf, unsigned rx_bufsize,
	struct ucg_cirbuf *tx_cbuf, char *tx_buf, unsigned tx_bufsize)
{
	const struct ucg_uart_config def_conf = {
		.enable = 1,
		.parity = 0,
		.stop_bits = 0,
		.reserved = 0,
		.nbits = 8,
		.baudrate = 9600,
	};

	memset(uart, 0, sizeof(*uart));
	ucg_cirbuf_init(rx_cbuf, rx_buf, 0, rx_bufsize);
	ucg_cirbuf_init(tx_cbuf, tx_buf, 0, tx_bufsize);
	uart->rx_fifo = rx_cbuf;
	uart->tx_fifo = tx_cbuf;
	uart->ops = ops;
	uart->driver_data = driver_data;
	ucg_uart_setconf(uart, &def_conf);
}
