#
# Copyright 2015, Olivier MATZ <zer0@droids-corp.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of California, Berkeley nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# convert source to obj file
src2obj = $(strip $(patsubst %.c,%.o,\
	$(patsubst %.s,%.o,$(1))))

# add a dot in front of the file name
dotfile = $(strip $(foreach f,$(1),\
	$(join $(dir $f),.$(notdir $f))))

# convert source/obj files into dot-dep filename
file2dep = $(strip $(call dotfile,$(addsuffix .d,$(1))))
file2tmpdep = $(strip $(call dotfile,$(addsuffix .d.tmp,$(1))))
file2cmd = $(strip $(call dotfile,$(addsuffix .cmd,$(1))))

OBJS := $(call src2obj,$(SRCS))
DEPS := $(call file2dep,$(OBJS))
CMDS := $(call file2cmd,$(OBJS))

# add a \ before each quote
protect_quote = $(subst ','\'',$(1))
#'# editor syntax highlight fix

# return 1 if parameter is a non-empty string, else 0
boolean = $(if $1,1,0)

# return an empty string if string are equal
compare = $(strip $(subst $(1),,$(2)) $(subst $(2),,$(1)))

# return a non-empty string if a file does not exist
#  $1: file
file_missing = $(call compare,$(wildcard $1),$1)

# return a non-empty string if cmdline changed
#  $1: file to be built
#  $2: the command to build it
cmdline_changed = $(call compare,$(strip $(cmd_$(1))),$(strip $(2)))

# return an non-empty string if the .d file does not exist
#  $1: the dep file (.d)
depfile_missing = $(call compare,$(wildcard $(1)),$(1))

# return a non-empty string if, according to dep_xyz variable, a file
# needed to build $1 does not exist. In this case we need to rebuild
# the file and the .d file.
#  $1: file to be built
dep_missing = $(call compare,$(wildcard $(dep_$(1))),$(dep_$(1)))

# return an empty string if no prereq is newer than target
#  $1: list of prerequisites ($^)
#  $2: list of prerequisites newer than target ($?)
dep_newer = $(strip $(filter-out FORCE,$(2)))

# display why a file should be re-built
#  $1: source files
#  $2: dst file
#  $3: build command
#  $4: all prerequisites ($^)
#  $5: all prerequisites newer than target ($?)
ifeq ($(D),1)
display_deps = \
	echo -n "$1 -> $2 " ; \
	echo -n "file_missing=$(call boolean,$(call file_missing,$(2))) " ; \
	echo -n "cmdline_changed=$(call boolean,$(call cmdline_changed,$(2),$(3))) " ; \
	echo -n "depfile_missing=$(call boolean,$(call depfile_missing,$(call file2dep,$(2)))) " ; \
	echo -n "dep_missing=$(call boolean,$(call dep_missing,$(2))) " ; \
	echo "dep_newer=$(call boolean,$(call dep_newer,$(4),$(5)))"
else
display_deps=
endif

# return an empty string if a file should be rebuilt
#  $1: dst file
#  $2: build command
#  $3: all prerequisites ($^)
#  $4: all prerequisites newer than target ($?)
check_deps = $(or $(call file_missing,$(1)),\
	$(call cmdline_changed,$(1),$(2)),\
	$(call depfile_missing,$(call file2dep,$(1))),\
	$(call dep_missing,$(1)),\
	$(call dep_newer,$(3),$(4)))

# fix the format of .o.d.tmp (generated by gcc) to a .o.d that defines
# dependencies as makefile variables
#  $1: object file (.o)
fixdep = if [ -f $(call file2tmpdep,$(1)) ]; then\
	echo -n "dep_$(abspath $(1)) = " > $(call file2dep,$(1)) && \
	sed 's,^[^ ][^:]*: ,,' $(call file2tmpdep,$(1)) >> $(call file2dep,$(1)) && \
	rm -f $(call file2tmpdep,$(1)); \
fi

# create a depfile with no additional deps
#  $1: file to be built
create_empty_depfile = echo "dep_$(abspath $(1)) =" > $(call file2dep,$(1))

# save a command in a file
#  $1: command to build the file
#  $2: name of the file
save_cmd = echo "cmd_$(2) = $(call protect_quote,$(1))" > $(call file2cmd,$(2))

# compile a file
#  $1: sources
#  $2: dst
compile = $(CC) -Wp,-MD,$(call file2tmpdep,$(2)) \
	$(CPPFLAGS) $(CFLAGS) -c -o $2 $1

# print line used to compile a file
ifeq ($(V),1)
compile_print_cmd = echo $(call protect_quote,$(call compile,$1,$2))
else
compile_print_cmd = echo "  CC $(2)"
endif

